'''Query and manipule data from any source.'''

import io
import os
import re
import time
import json
import sqlalchemy as sa
import sqlite3
import numpy as np
import pandas as pd
import gramex.cache
from datetime import datetime
from packaging import version
from tornado.escape import json_encode
from typing import Callable, List, Tuple, Dict, Union, Any
from gramex.config import merge, app_log
from gramex.transforms import build_transform
from orderedattrdict import AttrDict
from urllib.parse import urlparse

_ENGINE_CACHE = {}
_METADATA_CACHE = {}
_FOLDER = os.path.dirname(os.path.abspath(__file__))
# Dummy path used by _path_safe to detect sub-directories
_path_safe_root = os.path.realpath('/root/dir')
# Aggregator separator. ?col|SUM treats SUM as an aggregate function
_agg_sep = '|'
# List of aggregated types returned by operators (if different from column type)
# Note: For aggregation functions, see:
# SQLite: https://www.sqlite.org/lang_aggfunc.html
# MySQL: https://dev.mysql.com/doc/refman/8.0/en/group-by-functions.html
# PostgreSQL: https://www.postgresql.org/docs/9.5/static/functions-aggregate.html
# SQL Server: http://bit.ly/2MYPgQi
# DB2: https://ibm.co/2Kfbnjw
# Oracle: https://docs.oracle.com/database/121/SQLRF/functions003.htm
_agg_type = {
    'sum': float,
    'count': int,
    'avg': float,
    'stdev': float,  # MS SQL version of stddev
    'stddev': float,
    'rank': int,
    'percent_rank': float,
    # The following types are the same as the columns
    # first, last, min, max, median
}
# List of Python types returned by SQLAlchemy
_numeric_types = {'int', 'long', 'float', 'Decimal'}
# In SQLite, serialize Pandas timestamp exactly the way it serializes datetimes.
# https://github.com/python/cpython/blob/3.8/Lib/sqlite3/dbapi2.py#L66
sqlite3.register_adapter(pd.Timestamp, lambda d: d.isoformat(' '))
# Data processing plugins.
# e.g. plugins['mongodb'] = {'filter': fn, 'insert': fn, ...}
plugins = {}


def filter(
    url: Union[str, pd.DataFrame],
    args: dict = {},
    meta: dict = {},
    engine: str = None,
    table: str = None,
    ext: str = None,
    id: List[str] = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    transform: Callable = None,
    transform_kwargs: dict = {},
    argstype: Dict[str, dict] = {},
    **kwargs: dict,
) -> pd.DataFrame:
    '''Filter data using URL query parameters.

    Examples:
        >>> gramex.data.filter(dataframe, args=handler.args)
        >>> gramex.data.filter('file.csv', args=handler.args)
        >>> gramex.data.filter('mysql://server/db', table='table', args={'user': [user']})

    Parameters:

        url: DataFrame, sqlalchemy URL, directory or file name, http(s) URL
        args: URL query parameters as a dict of lists. Pass handler.args or parse_qs results
        meta: this dict is updated with metadata during the course of filtering
        engine: over-rides the auto-detected engine. Can be 'dataframe', 'file',
            'http', 'https', 'sqlalchemy', 'dir'
        ext: file extension (if url is a file). Defaults to url extension
        columns: database column names to create if required (if url is a database).
            Keys are column names. Values can be SQL types, or dicts with these keys:
                - `type` (str), e.g. `"VARCHAR(10)"`
                - `default` (str/int/float/bool), e.g. `"none@example.org"`
                - `nullable` (bool), e.g. `False`
                - `primary_key` (bool), e.g. `True` -- used only when creating new tables
                - `autoincrement` (bool), e.g. `True` -- used only when creating new tables
        query: optional SQL query to execute (if url is a database),
            `.format`-ed using `args` and supports SQLAlchemy SQL parameters.
            Loads entire result in memory before filtering.
        queryfile: optional SQL query file to execute (if url is a database).
            Same as specifying the `query:` in a file. Overrides `query:`
        transform: optional in-memory transform of source data. Takes
            the result of gramex.cache.open or gramex.cache.query. Must return a
            DataFrame. Applied to both file and SQLAlchemy urls.
        transform_kwargs: optional keyword arguments to be passed to the
            transform function -- apart from data
        argstype: optional dict that specifies `args` type and behavior for `query`.
        **kwargs: Additional parameters are passed to
            [gramex.cache.open][], `sqlalchemy.create_engine` or the plugin's filter

    Returns:
        Filtered DataFrame

    To filter a DataFrame where column x=1 and y=2:

        >>> filtered = gramex.data.filter(dataframe, args={'x': [1], 'y': [2]})

    `args` is always a dict of lists, which is compatible with Gramex's `handler.args`.
    So you can replace the above with:

        >>> filtered = gramex.data.filter(dataframe, args=handle.args)

    To filter ?city=Rome from a CSV/XLS/any file supported by [gramex.cache.open][]:

        >>> gramex.data.filter('path/to/file.csv', rel=True, args={'city': ['Rome']})

    Remaining `kwargs` are passed to [gramex.cache.open][] if `url` is a file. So `rel=True` works

    To filter ?city=Rome from a SQLite, MySQL, PostgreSQL or any SQLAlchemy-supported DB:

        >>> gramex.data.filter('sqlite:///x.db', table='data', args={'city': ['Rome']})

    Remaining `kwargs` are passed to `sqlalchemy.create_engine` if `url` is a SQLAlchemy URL. E.g.

    - `table`: table name (if url is an SQLAlchemy URL), `.format`-ed using `args`.
    - `state`: optional SQL query to check if data has changed.

    TODO: Document how to pass params -- for each database

    If `table` or `query` is passed to an SQLAlchemy url, it is formatted using `args`.
    For example

        >>> data = gramex.data.filter('mysql://server/db', table='{xxx}', args=handler.args)

    ... when passed `?xxx=sales` returns rows from the sales table. Similarly:

        >>> data = gramex.data.filter(
        ...     'mysql://server/db', args=handler.args,
        ...     query='SELECT {col}, COUNT(*) FROM table GROUP BY {col}')

    ... when passsed `?col=City` replaces `{col}` with `City`.

    NOTE: To avoid SQL injection attacks, only keys without spaces are allowed.
    So `?city name=Oslo` **will not** work.

    The URL supports operators filter like this:

    - `?x` selects x is not null
    - `?x!` selects x is null
    - `?x=val` selects x == val
    - `?x!=val` selects x != val
    - `?x>=val` selects x > val
    - `?x>~=val` selects x >= val
    - `?x<=val` selects x < val
    - `?x<~=val` selects x <= val
    - `?x~=val` selects x matches val
    - `?x!~=val` selects x does not match val
    - `?x*=val` selects x matches val case-insensitively
    - `?x!*=val` selects x does not match val case-insensitively

    Multiple filters are combined into an AND clause. Ranges can also be
    specified like this:

    - `?x=a&y=b` selects x = a AND y = b
    - `?x>=100&x<=200` selects x > 100 AND x < 200

    If the same column has multiple values, they are combined like this:

    - `?x=a&x=b` selects x IN (a, b)
    - `?x!=a&x!=b` selects x NOT IN (a, b)
    - `?x~=a&x~=b` selects x ~ a|b
    - `?x>=a&x>=b` selects x > MIN(a, b)
    - `?x<=a&x<=b` selects x < MAX(a, b)

    Arguments are converted to the type of the column before comparing. If this
    fails, it raises a ValueError.

    You can specify the SQL argument type for database queries with `argstype`. This (currently)
    only applies when using `query` or `queryfunction`, not `table`:

    - `argstype: {'x': {type: float}, 'y': {type: bool}}` treats x as a float and y as a bool
    - `argstype: {'x': {type: int, expanding=True}}` treats x as a list of int, suitable for
      use in an `IN` clause, e.g. `SELECT * FROM table WHERE size IN :x`

    These URL query parameters control the output:

    - `?_sort=col` sorts column col in ascending order. `?_sort=-col` sorts
        in descending order.
    - `?_limit=100` limits the result to 100 rows
    - `?_offset=100` starts showing the result from row 100. Default: 0
    - `?_c=x&_c=y` returns only columns `[x, y]`. `?_c=-col` drops col.

    If a column name matches one of the above, you cannot filter by that column.
    Avoid column names beginning with _.

    To get additional information about the filtering, use:

        meta = {}  # Create a variable which will be filled with more info
        filtered = gramex.data.filter(data, meta=meta, **handler.args)

    The `meta` variable is populated with the following keys:

    - `filters`: Applied filters as `[(col, op, val), ...]`
    - `ignored`: Ignored filters as `[(col, vals), ('_sort', col), ('_by', col), ...]`
    - `excluded`: Excluded columns as `[col, ...]`
    - `sort`: Sorted columns as `[(col, True), ...]`. The second parameter is `ascending=`
    - `offset`: Offset as integer. Defaults to 0
    - `limit`: Limit as integer - `None` if limit is not applied
    - `count`: Total number of rows, if available
    - `by`: Group by columns as `[col, ...]`
    - `inserted`: List of (dict of primary values) for each inserted row

    These variables may be useful to show additional information about the
    filtered data.
    '''
    # Auto-detect engine.
    if engine is None:
        engine = get_engine(url)

    # Pass the meta= argument from kwargs (if any)
    meta.update(
        {
            'filters': [],  # Applied filters as [(col, op, val), ...]
            'ignored': [],  # Ignored filters as [(col, vals), ...]
            'sort': [],  # Sorted columns as [(col, asc), ...]
            'offset': 0,  # Offset as integer
            'limit': None,  # Limit as integer - None if not applied
            'by': [],  # Group by columns as [col, ...]
        }
    )
    args = dict(args)  # Do not modify the args -- keep a copy
    controls = _pop_controls(args)
    transform = _transform_fn(transform, transform_kwargs)
    url, ext, query, queryfile, table, kwargs = _replace(
        engine, args, url, ext, query, queryfile, table, **kwargs
    )

    # Use the appropriate filter function based on the engine
    if engine == 'dataframe':
        data = transform(url) if callable(transform) else url
        return _filter_frame(data, meta, controls, args, argstype)
    elif engine == 'dir':
        data = dirstat(url, **args)
        data = transform(data) if callable(transform) else data
        return _filter_frame(data, meta, controls, args, argstype)
    elif engine in {'file', 'http', 'https'}:
        if engine == 'file' and not os.path.exists(url):
            raise OSError(f'url: {url} not found')
        # table= is not a valid option for all gramex.cache.open formats. Use only if specified
        if table is not None:
            kwargs['table'] = table
        # Get the full dataset. Then filter it
        data = gramex.cache.open(url, ext, transform=transform, **kwargs)
        return _filter_frame(data, meta, controls, args, argstype)
    elif engine.startswith('plugin+'):
        plugin = engine.split('+')[1]
        method = plugins[plugin]['filter']
        return method(
            url=url,
            meta=meta,
            controls=controls,
            args=args,
            argstype=argstype,
            id=id,
            table=table,
            columns=columns,
            ext=ext,
            query=query,
            queryfile=queryfile,
            **kwargs,
        )
    elif engine == 'sqlalchemy':
        state = kwargs.pop('state', None)
        engine = alter(url, table, columns, **kwargs)
        if query or queryfile:
            if queryfile:
                query = gramex.cache.open(queryfile, 'text')
            if not state:
                if isinstance(table, str):
                    state = table if ' ' in table else [table]
                elif isinstance(table, (list, tuple)):
                    state = list(table)
                elif table is not None:
                    raise ValueError(f'table: must be string or list of strings, not {table!r}')
            # sa.text() provides backend-neutral :name for bind parameters
            # NOTE: sa.text() caches queries => .bindparams() UPDATES previous bindparams. So:
            #   If query = "SELECT * FROM table WHERE x=:x" and we bind with bindparam('x'),
            #   we can NEVER bind with x as bindparam('x', expanding=True) without restarting.
            #   Therefore, create a new query each time.
            sql = sa.text(f'{query.rstrip().rstrip(";")}; -- {time.time()}')
            all_params = {}
            for key, vals in args.items():
                conv, expanding = _argstype(argstype, key, str)
                if expanding:
                    all_params[key] = tuple(conv(val) for val in vals if val)
                    sql = sql.bindparams(sa.bindparam(key, expanding=True))
                elif len(vals) > 0:
                    all_params[key] = conv(vals[0])
            data = gramex.cache.query(sql, engine, state, params=all_params)
            data = transform(data) if callable(transform) else data
            # The query acts as base data. Now filter with additional parameters
            return _filter_frame(data, meta, controls, args, argstype)
        elif table:
            if callable(transform):
                data = gramex.cache.query(table, engine, [table])
                return _filter_frame(transform(data), meta, controls, args, argstype)
            else:
                return _filter_db(engine, table, meta, controls, args, argstype)
        else:
            raise ValueError('No table: or query: specified')
    else:
        raise ValueError(f'engine: {engine} invalid. Can be sqlalchemy|file|dataframe')


def delete(
    url: Union[str, pd.DataFrame],
    args: dict = {},
    meta: dict = {},
    engine: str = None,
    table: str = None,
    ext: str = None,
    id: List[str] = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    transform: Callable = None,
    transform_kwargs: dict = {},
    argstype: Dict[str, dict] = {},
    **kwargs: dict,
) -> int:
    '''Deletes data using URL query parameters.

    Examples:
        >>> gramex.data.delete(dataframe, args={'city': ['Oslo']})
        >>> gramex.data.delete('mysql://server/db', table='x', args={'city': ['Oslo']})
        >>> gramex.data.delete(url, args=handler.args)

    It accepts the same parameters as [gramex.data.filter][], and returns the number
    of deleted rows.
    '''
    if engine is None:
        engine = get_engine(url)
    meta.update({'filters': [], 'ignored': []})
    args = dict(args)  # Do not modify the args -- keep a copy
    controls = _pop_controls(args)
    url, table, ext, query, queryfile, kwargs = _replace(
        engine, args, url, table, ext, query, queryfile, **kwargs
    )
    if engine == 'dataframe':
        data_filtered = _filter_frame(url, meta, controls, args, argstype, source='delete', id=id)
        return len(data_filtered)
    elif engine == 'file':
        data = gramex.cache.open(url, ext, transform=transform, **kwargs)
        data_filtered = _filter_frame(data, meta, controls, args, argstype, source='delete', id=id)
        gramex.cache.save(data, url, ext, index=False, **kwargs)
        return len(data_filtered)
    elif engine.startswith('plugin+'):
        plugin = engine.split('+')[1]
        method = plugins[plugin]['delete']
        return method(
            url=url,
            meta=meta,
            controls=controls,
            args=args,
            argstype=argstype,
            id=id,
            table=table,
            columns=columns,
            ext=ext,
            query=query,
            queryfile=queryfile,
            **kwargs,
        )
    elif engine == 'sqlalchemy':
        if table is None:
            raise ValueError('No table: specified')
        engine = alter(url, table, columns, **kwargs)
        return _filter_db(engine, table, meta, controls, args, argstype, source='delete', id=id)
    else:
        raise ValueError(f'engine: {engine} invalid. Can be sqlalchemy|file|dataframe')


def update(
    url: Union[str, pd.DataFrame],
    args: dict = {},
    meta: dict = {},
    engine: str = None,
    table: str = None,
    ext: str = None,
    id: List[str] = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    transform: Callable = None,
    transform_kwargs: dict = {},
    argstype: Dict[str, dict] = {},
    **kwargs: dict,
) -> int:
    '''Update data using URL query parameters.

    Examples:
        >>> gramex.data.update(dataframe, args=handler.args, id=['id'])
        >>> gramex.data.update('file.csv', args=handler.args, id=['id'])
        >>> gramex.data.update('mysql://server/db', table='x', args=handler.args, id=['id'])

    `id` is a list of column names defining the primary key.
    Calling this in a handler with `?id=1&x=2` updates x=2 where id=1.

    It accepts the same parameters as [gramex.data.filter][], and returns the number of updated
    rows.
    '''
    if engine is None:
        engine = get_engine(url)
    meta.update({'filters': [], 'ignored': []})
    args = dict(args)  # Do not modify the args -- keep a copy
    controls = _pop_controls(args)
    url, table, ext, query, queryfile, kwargs = _replace(
        engine, args, url, table, ext, query, queryfile, **kwargs
    )
    if engine == 'dataframe':
        data_updated = _filter_frame(url, meta, controls, args, argstype, source='update', id=id)
        return len(data_updated)
    elif engine == 'file':
        data = gramex.cache.open(url, ext, transform=transform, **kwargs)
        data_updated = _filter_frame(data, meta, controls, args, argstype, source='update', id=id)
        gramex.cache.save(data, url, ext, index=False, **kwargs)
        return len(data_updated)
    elif engine.startswith('plugin+'):
        plugin = engine.split('+')[1]
        method = plugins[plugin]['update']
        return method(
            url=url,
            meta=meta,
            controls=controls,
            args=args,
            argstype=argstype,
            id=id,
            table=table,
            columns=columns,
            ext=ext,
            query=query,
            queryfile=queryfile,
            **kwargs,
        )
    elif engine == 'sqlalchemy':
        if table is None:
            raise ValueError('No table: specified')
        engine = alter(url, table, columns, **kwargs)
        return _filter_db(engine, table, meta, controls, args, argstype, source='update', id=id)
    else:
        raise ValueError(f'engine: {engine} invalid. Can be sqlalchemy|file|dataframe')


def insert(
    url: Union[str, pd.DataFrame],
    args: dict = {},
    meta: dict = {},
    engine: str = None,
    table: str = None,
    ext: str = None,
    id: List[str] = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    transform: Callable = None,
    transform_kwargs: dict = {},
    argstype: Dict[str, dict] = {},
    **kwargs: dict,
) -> int:
    '''Insert data using URL query parameters.

    Examples:
        >>> gramex.data.insert(dataframe, args=handler.args, id=['id'])
        >>> gramex.data.insert('file.csv', args=handler.args, id=['id'])
        >>> gramex.data.insert('mysql://server/db', table='x', args=handler.args, id=['id'])

    `id` is a list of column names defining the primary key.
    Calling this in a handler with `?id=3&x=2` inserts a new record with id=3 and x=2.

    If the target file / table does not exist, it is created.

    It accepts the same parameters as [gramex.data.filter][], and returns the number of updated
    rows.
    '''
    if engine is None:
        engine = get_engine(url)
    args = dict(args)  # Do not modify the args -- keep a copy
    controls = _pop_controls(args)
    if not args:
        raise ValueError('No args: specified')
    meta.update({'filters': [], 'ignored': [], 'inserted': []})
    # If values do not have equal number of elements, pad them and warn
    rowcount = max(len(val) for val in args.values())
    for key, val in args.items():
        rows = len(val)
        if 0 < rows < rowcount:
            val += [val[-1]] * (rowcount - rows)
            app_log.warning(
                f'data.insert: column {key} has {rows} rows not {rowcount}. '
                f'Extended last value {val[-1]}'
            )
    rows = pd.DataFrame.from_dict(args)
    url, table, ext, query, queryfile, kwargs = _replace(
        engine, args, url, table, ext, query, queryfile, **kwargs
    )
    if engine == 'dataframe':
        rows = _pop_columns(rows, url.columns, meta['ignored'])
        url = pd.concat([url, pd.DataFrame.from_records(rows)])
        return len(rows)
    elif engine == 'file':
        try:
            data = gramex.cache.open(url, ext, transform=None, **kwargs)
        except OSError:
            data = rows
        else:
            rows = _pop_columns(rows, data.columns, meta['ignored'])
            data = pd.concat([data, pd.DataFrame.from_records(rows)])
        gramex.cache.save(data, url, ext, index=False, **kwargs)
        return len(rows)
    elif engine.startswith('plugin+'):
        plugin = engine.split('+')[1]
        method = plugins[plugin]['insert']
        return method(
            url=url,
            meta=meta,
            controls=controls,
            rows=rows,
            args=args,
            argstype=argstype,
            id=id,
            table=table,
            columns=columns,
            ext=ext,
            query=query,
            queryfile=queryfile,
            **kwargs,
        )
    elif engine == 'sqlalchemy':
        if table is None:
            raise ValueError('No table: specified')
        engine = alter(url, table, columns, **kwargs)
        try:
            cols = get_table(engine, table).columns
        except sa.exc.NoSuchTableError:
            pass
        else:
            rows = _pop_columns(rows, [col.name for col in cols], meta['ignored'])
        if '.' in table:
            kwargs['schema'], table = table.rsplit('.', 1)
        # SQLAlchemy 1.4+ only allows sa.inspect() for all table introspection
        if version.parse(sa.__version__) >= version.parse('1.4'):
            has_table = sa.inspect(engine).has_table(table, schema=kwargs.get('schema'))
        # SQLAlchemy 1.3- does not have .has_table()
        else:
            has_table = engine.dialect.has_table(engine, table, schema=kwargs.get('schema'))
        # If the DB doesn't yet have the table, create it WITH THE PRIMARY KEYS.
        if not has_table and id:
            # Note: pandas does not document get_schema, so it might change.
            engine.execute(pd.io.sql.get_schema(rows, name=table, keys=id, con=engine))

        def insert_method(tbl, conn, keys, data_iter):
            '''Pandas .to_sql() doesn't return inserted row primary keys. Capture it in meta'''
            data = [dict(zip(keys, row)) for row in data_iter]
            # If the ?id= is not provided, Pandas creates a schema based on available columns,
            # without the `id` column. SQLAlchemy won't return inserted_primary_key unless the
            # metadata has a primary key. So, hoping that the table already has a primary key,
            # load table from DB via extend_existing=True.
            sa_table = sa.Table(
                table, tbl.table.metadata, extend_existing=True, autoload_with=engine
            )
            r = conn.execute(sa_table.insert(), data)
            # SQLAlchemy 1.4+ supports inserted_primary_key_rows.
            if hasattr(r, 'inserted_primary_key_rows'):
                ids = r.inserted_primary_key_rows
            # In SQLAlchemy 1.3, only single inserts have an inserted_primary_key.
            elif hasattr(r, 'inserted_primary_key'):
                ids = [r.inserted_primary_key]
            else:
                ids = []
            # Add non-empty IDs as a dict with associated keys.
            # If there are no auto-generated primary keys in the table, no need to return anything.
            id_cols = [col.name for col in sa_table.primary_key]
            for row in ids:
                if row:
                    meta['inserted'].append(dict(zip(id_cols, row)))

        kwargs['method'] = insert_method
        # If user passes ?col= with an empty string, replace with NULL;
        # because, if the column is an INT/FLOAT, type conversion int('') / float('') will fail.
        for col in rows.columns:
            if rows[col].dtype == object:
                rows[col].replace({'': None}, inplace=True)

        # kwargs might contain additonal unexpected values, pass expected arguments explicitly
        pd.io.sql.to_sql(
            rows,
            table,
            engine,
            if_exists='append',
            index=False,
            schema=kwargs.get('schema'),
            index_label=kwargs.get('index_label'),
            chunksize=kwargs.get('chunksize'),
            dtype=kwargs.get('dtype'),
            method=kwargs.get('method'),
        )

        return len(rows)
    else:
        raise ValueError(f'engine: {engine} invalid. Can be sqlalchemy|file|dataframe')


def get_engine(url: Union[str, pd.DataFrame]) -> str:
    '''Detect the type of url passed.

    The return value is

    - `'dataframe'` if url is a Pandas DataFrame
    - `'sqlalchemy'` if url is a sqlalchemy compatible URL
    - `'plugin'` if it is `<valid-plugin-name>://...`
    - `protocol` if url is of the form `protocol://...`
    - `'dir'` if it is not a URL but a valid directory
    - `'file'` if it is not a URL but a valid file
    - `None` otherwise
    '''
    if isinstance(url, pd.DataFrame):
        return 'dataframe'
    for plugin_name in plugins:
        if url.startswith(f'{plugin_name}:'):
            return f'plugin+{plugin_name}'
    try:
        url = sa.engine.url.make_url(url)
    except sa.exc.ArgumentError:
        return 'dir' if os.path.isdir(url) else 'file'
    try:
        url.get_driver_name()
        return 'sqlalchemy'
    except sa.exc.NoSuchModuleError:
        return url.drivername


def create_engine(url: str, create: sa.engine.base.Engine = sa.create_engine, **kwargs: dict):
    '''
    Cached version of sqlalchemy.create_engine (or any custom engine).

    Normally, this is not required. But [gramex.data.get_table][] caches the engine
    *and* metadata *and* uses autoload=True. This makes sqlalchemy create a new
    database connection for every engine object, and not dispose it. So we
    re-use the engine objects within this module.
    '''
    if url not in _ENGINE_CACHE:
        _ENGINE_CACHE[url] = create(url, **kwargs)
    return _ENGINE_CACHE[url]


def get_table(engine: sa.engine.base.Engine, table: str, **kwargs: dict) -> sa.Table:
    '''Return the sqlalchemy table from the engine and table name'''
    if engine not in _METADATA_CACHE:
        _METADATA_CACHE[engine] = sa.MetaData(bind=engine)
    metadata = _METADATA_CACHE[engine]
    if '.' in table:
        kwargs['schema'], table = table.rsplit('.', 1)
    return sa.Table(table, metadata, autoload=True, autoload_with=engine, **kwargs)


def download(
    data: Union[pd.DataFrame, List[pd.DataFrame]],
    format: str = 'json',
    template: str = None,
    args: dict = {},
    **kwargs: dict,
) -> bytes:
    '''
    Download a DataFrame or dict of DataFrames in various formats. This is used
    by [gramex.handlers.FormHandler][]. You are **strongly** advised to
    try it before creating your own FunctionHandler.

    Usage as a FunctionHandler:

        def download_as_csv(handler):
            handler.set_header('Content-Type', 'text/csv')
            handler.set_header('Content-Disposition', 'attachment;filename=data.csv')
            return gramex.data.download(dataframe, format='csv')

    It takes the following arguments:

    Parameters:
        data: A DataFrame or a dict of DataFrames
        format: Output format. Can be `csv|json|html|xlsx|template`
        template: Path to template file for `template` format
        args: dictionary of user arguments to subsitute spec
        **kwargs: Additional parameters that are passed to the relevant renderer

    Returns:
        bytes with the download file contents

    When `data` is a DataFrame, this is what different `format=` parameters
    return:

    - `csv` returns a UTF-8-BOM encoded CSV file of the dataframe
    - `xlsx` returns an Excel file with 1 sheet named `data`. kwargs are
        passed to `.to_excel(index=False)`
    - `html` returns a HTML file with a single table. kwargs are passed to
        `.to_html(index=False)`
    - `json` returns a JSON file. kwargs are passed to
        `.to_json(orient='records', force_ascii=True)`.
    - `template` returns a Tornado template rendered file. The template
        receives `data` as `data` and any additional kwargs.
    - `pptx` returns a PPTX generated by pptgen
    - `seaborn` or `sns` returns a Seaborn generated chart
    - `vega` returns JavaScript that renders a Vega chart

    When `data` is a dict of DataFrames, the following additionally happens:

    - `format='csv'` renders all DataFrames one below the other, adding the
        key as heading
    - `format='xlsx'` renders each DataFrame on a sheet whose name is the key
    - `format='html'` renders tables below one another with the key as heading
    - `format='json'` renders as a dict of DataFrame JSONs
    - `format='template'` sends `data` and all `kwargs` as passed to the
        template
    - `format='pptx'` passes `data` as a dict of datasets to pptgen
    - `format='vega'` passes `data` as a dict of datasets to Vega

    When `data` is NEITHER a DataFrame or a dict of DataFrames:

    - `format='json'` renders as JSON if possible
    - `format='template'` renders the template if possible
    - all other formats raise a `ValueError`

    You need to set the MIME types on the handler yourself. Recommended MIME
    types are in gramex.yaml under handler.FormHandler.
    '''
    multiple_datasets = True
    error_no_dataframe = None

    # Check if data is a DataFrame or a dict of DataFrames (multiple_datasets).
    # Ensure that data becomes a dict of DataFrames
    if isinstance(data, dict):
        for key, val in data.items():
            if not isinstance(val, pd.DataFrame):
                error_no_dataframe = f'download(): {key} type is {type(val)}, not a DataFrame'
        if not len(data):
            error_no_dataframe = 'download(): got empty dict. Need a DataFrame'
    elif not isinstance(data, pd.DataFrame):
        error_no_dataframe = f'download(): type is {type(data)}, not a DataFrame'
    else:
        data = {'data': data}
        multiple_datasets = False

    # These formats require a DataFrame or a dict of DataFrames. Other formats (json, template)
    # accept anything.
    if error_no_dataframe and format in {
        'csv',
        'html',
        'xlsx',
        'xls',
        'pptx',
        'ppt',
        'seaborn',
        'sns',
        'vega',
        'vega-lite',
        'vegam',
    }:
        raise ValueError(error_no_dataframe)

    def kw(**conf):
        '''Set provided conf as defaults for kwargs'''
        return merge(kwargs, conf, mode='setdefault')

    if format == 'csv':
        # csv.writer requires BytesIO in PY2 and StringIO in PY3.
        # I can't see an elegant way out of this other than writing code for each.
        out = io.StringIO()
        kw(index=False)
        for index, (key, val) in enumerate(data.items()):
            if index > 0:
                out.write('\n')
            if multiple_datasets:
                out.write(key + '\n')
            val.to_csv(out, **kwargs)
        result = out.getvalue()
        # utf-8-sig encoding returns the result with a UTF-8 BOM. Easier to open in Excel
        return result.encode('utf-8-sig') if result.strip() else result.encode('utf-8')
    elif format == 'template':
        return gramex.cache.open(template, 'template').generate(
            data=data if multiple_datasets else data['data'], **kwargs
        )
    elif format == 'html':
        out = io.StringIO()
        kw(index=False)
        for key, val in data.items():
            if multiple_datasets:
                out.write(f'<h1>{key}</h1>')
            val.to_html(out, **kwargs)
        return out.getvalue().encode('utf-8')
    elif format in {'xlsx', 'xls'}:
        out = io.BytesIO()
        kw(index=False)
        # TODO: Create and use a FrameWriter for formatting
        with pd.ExcelWriter(out, engine='xlsxwriter') as writer:
            for key, val in data.items():
                val.to_excel(writer, sheet_name=key, **kwargs)
        return out.getvalue()
    elif format in {'pptx', 'ppt'}:
        from gramex.pptgen import pptgen

        kw()
        out = io.BytesIO()
        pptgen(target=out, data=data, **kwargs)
        return out.getvalue()
    elif format in {'seaborn', 'sns'}:
        kw = AttrDict()
        defaults = {
            'chart': 'barplot',
            'ext': 'png',
            'data': 'data',
            'dpi': 96,
            'width': 640,
            'height': 480,
        }
        for key, default in defaults.items():
            kw[key] = kwargs.pop(key, default)
        import matplotlib

        matplotlib.use('Agg')  # Before importing seaborn, set a headless backend
        import seaborn as sns

        plot = getattr(sns, kw.chart)(data=data.get(kw.data), **kwargs)
        out = io.BytesIO()
        fig = plot.figure if hasattr(plot, 'figure') else plot.fig
        for k in ['dpi', 'width', 'height']:
            kw[k] = float(kw[k])
        fig.set_size_inches(kw.width / kw.dpi, kw.height / kw.dpi)
        fig.savefig(out, format=kw.ext, dpi=kw.dpi)
        fig.clear()
        return out.getvalue()
    elif format in {'vega', 'vega-lite', 'vegam'}:
        kwargs = kw(orient='records', force_ascii=True)
        spec = kwargs.pop('spec', {})
        kwargs.pop('handler', None)
        out = io.BytesIO()
        # conf = {..., spec: {..., data: __DATA__}}
        if isinstance(spec.get('data'), (dict, list)) or 'fromjson' in spec:
            # support only one dataset
            values = list(data.values())
            out.write(values[0].to_json(**kwargs).encode('utf-8'))
            out = out.getvalue()
        else:
            spec['data'] = '__DATA__'
            for index, (key, val) in enumerate(data.items()):
                out.write(b',{"name":' if index > 0 else b'{"name":')
                out.write(json_encode(key).encode('utf-8'))
                out.write(b',"values":')
                out.write(val.to_json(**kwargs).encode('utf-8'))
                out.write(b'}')
            out = out.getvalue()
            if format == 'vega':
                out = b'[' + out + b']'
        kwargs['spec'], _ = _replace('', args, spec)
        conf = json.dumps(kwargs, ensure_ascii=True, separators=(',', ':'), indent=None)
        conf = conf.encode('utf-8').replace(b'"__DATA__"', out)
        script = gramex.cache.open(_VEGA_SCRIPT, 'bin')
        return script.replace(b'/*{conf}*/', conf)
    # If it's none of these formats, default to JSON.
    # If there are no DataFrames, handle arbitrary JSON
    elif error_no_dataframe:
        from gramex.config import CustomJSONEncoder

        kw(cls=CustomJSONEncoder)
        return json.dumps(data, **kwargs)
    # If there ARE DataFrames, render each
    else:
        out = io.BytesIO()
        kwargs = kw(orient='records', force_ascii=True)
        if multiple_datasets:
            out.write(b'{')
            for index, (key, val) in enumerate(data.items()):
                if index > 0:
                    out.write(b',')
                out.write(json_encode(key).encode('utf-8'))
                out.write(b':')
                out.write(val.to_json(**kwargs).encode('utf-8'))
            out.write(b'}')
        else:
            out.write(data['data'].to_json(**kwargs).encode('utf-8'))
        return out.getvalue()


def dirstat(url: str, timeout: int = 10, **kwargs: dict) -> pd.DataFrame:
    '''Return a DataFrame with the list of all files & directories under the url.

    Parameters:
        url: path to a directory, or a URL like `dir:///c:/path/`, `dir:////root/dir/`
        timeout: max seconds to wait. `None` to wait forever

    Raises:
        OSError: if url points to a missing location or is not a directory.

    Returns:
        DataFrame with 1 row per file/directory in the URL

    The result has these columns.

    - `type`: extension with a `.` prefix -- or `dir`
    - `dir`: directory path to the file relative to the URL
    - `name`: file name (including extension)
    - `path`: full path to file or dir. This equals url / dir / name
    - `size`: file size
    - `mtime`: last modified time in seconds since epoch
    - `level`: path depth (i.e. the number of paths in dir)
    '''
    target = url[7:] if url.startswith('dir:///') else url
    if not os.path.isdir(target):
        raise OSError(f'dirstat: {target} is not a directory')
    target = os.path.normpath(target)
    result = []
    start_time = time.time()
    for dirpath, dirnames, filenames in os.walk(target):
        if timeout and time.time() - start_time > timeout:
            app_log.debug(f'dirstat: {target} timeout ({timeout:.1f}s)')
            break
        for name in dirnames:
            path = os.path.join(dirpath, name)
            stat = os.stat(path)
            dirname = dirpath.replace(target, '').replace(os.sep, '/') + '/'
            result.append(
                {
                    'path': path,
                    'dir': dirname,
                    'name': name,
                    'type': 'dir',
                    'size': stat.st_size,
                    'mtime': stat.st_mtime,
                    'level': dirname.count('/'),
                }
            )
        for name in filenames:
            path = os.path.join(dirpath, name)
            stat = os.stat(path)
            dirname = dirpath.replace(target, '').replace(os.sep, '/') + '/'
            result.append(
                {
                    'path': path,
                    'dir': dirname,
                    'name': name,
                    'type': os.path.splitext(name)[-1],
                    'size': stat.st_size,
                    'mtime': stat.st_mtime,
                    'level': dirname.count('/'),
                }
            )
    return pd.DataFrame(result)


def filtercols(
    url: Union[str, pd.DataFrame],
    args: dict = {},
    meta: dict = {},
    engine: str = None,
    ext: str = None,
    query: str = None,
    queryfile: str = None,
    transform: Callable = None,
    transform_kwargs: dict = {},
    separator: str = ',',
    in_memory: bool = False,
    **kwargs: dict,
) -> pd.DataFrame:
    '''Filter data and extract unique values of each column using URL query parameters.

    Examples:
        >>> gramex.data.filtercols(dataframe, args=handler.args)
        >>> gramex.data.filtercols('file.csv', args=handler.args)
        >>> gramex.data.filtercols('mysql://server/db', table='table', args=handler.args)

    Parameters:

        url: Pandas DataFrame, sqlalchemy URL, directory or file name,
            `.format`-ed using `args`.
        args: URL query parameters as a dict of lists. Pass handler.args or parse_qs results
        meta: this dict is updated with metadata during the course of filtering
        engine: over-rides the auto-detected engine. Can be 'dataframe', 'file',
            'http', 'https', 'sqlalchemy', 'dir'
        ext: file extension (if url is a file). Defaults to url extension
        query: optional SQL query to execute (if url is a database),
            `.format`-ed using `args` and supports SQLAlchemy SQL parameters.
            Loads entire result in memory before filtering.
        queryfile: optional SQL query file to execute (if url is a database).
            Same as specifying the `query:` in a file. Overrides `query:`
        transform (function): optional in-memory transform of source data. Takes
            the result of gramex.cache.open or gramex.cache.query. Must return a
            DataFrame. Applied to both file and SQLAlchemy urls.
        transform_kwargs: optional keyword arguments to be passed to the
            transform function -- apart from data
        separator: string that separates columns in a hierarchy. Defaults to `,`.
            For example, `?_c=a,b` treats columns `a` and `b` as a tuple / hierarchy and
            filters them *together*.
        in_memory: fetch all unique values and compute filters in-memory. Faster,
            but takes more memory.
        **kwargs: Additional parameters are passed to
            [gramex.cache.open][] or `sqlalchemy.create_engine`

    Returns:
        filtered DataFrame

    Remaining kwargs are passed to [gramex.cache.open][] if `url` is a file, or
    `sqlalchemy.create_engine` if `url` is a SQLAlchemy URL.

    If this is used in a handler as

        >>> filtered = gramex.data.filtercols(dataframe, args=handler.args)

    ... then calling the handler with `?_c=state&_c=district` returns all unique values
    in columns of `dataframe` where columns are state and district.

    Column filter is supported like this:

    - `?_c=y&x` returns df with unique values of y where x is not null
    - `?_c=y&x=val` returns df with unique values of y where x == val
    - `?_c=y&y=val` returns df with unique values of y, ignores filter y == val
    - `?_c=y&x>=val` returns df with unique values of y where x > val
    - `?_c=x&_c=y&x=val` returns df with unique values of x ignoring filter x == val
        and returns unique values of y where x == val

    Arguments are converted to the type of the column before comparing. If this
    fails, it raises a ValueError.

    These URL query parameters control the output:

    - `?_sort=col` sorts column col in ascending order. `?_sort=-col` sorts
        in descending order.
    - `?_limit=100` limits the result to 100 rows
    - `?_offset=100` starts showing the result from row 100. Default: 0
    - `?_c=x&_c=y` returns only columns `[x, y]`. `?_c=-col` drops col.

    If a column name matches one of the above, you cannot filter by that column.
    Avoid column names beginning with _.

    You can handle hierarchies by passing `?_c=Country,City` with a comma. This returns all unique
    *combinations* of `Country` and `City`.

    To get the min/max or a column, use aggregations, e.g. `?_c=age|min&_c=age|max`.
    You can use `?_c=age|range` as a shortcut that returns min and max of a column.

    To get additional information about the filtering, use:

        meta = {}      # Create a variable which will be filled with more info
        filtered = gramex.data.filter(data, meta=meta, **handler.args)

    The `meta` variable is populated with the following keys:

    - `filters`: Applied filters as `[(col, op, val), ...]`
    - `ignored`: Ignored filters as `[(col, vals), ('_sort', cols), ...]`
    - `excluded`: Excluded columns as `[col, ...]`
    - `sort`: Sorted columns as `[(col, True), ...]`. The second parameter is `ascending=`
    - `offset`: Offset as integer. Defaults to 0
    - `limit`: Limit as integer - `100` if limit is not applied
    - `count`: Total number of rows, if available

    These variables may be useful to show additional information about the
    filtered data.
    '''
    # Auto-detect engine.
    if engine is None:
        engine = get_engine(url)
    result = {}
    limit = args.get('_limit', [100])
    try:
        limit = min(int(v) for v in limit)
    except ValueError:
        raise ValueError(f'_limit not integer: {limit!r}')
    if in_memory:
        # Fetch the superset data, i.e. all filter columns
        in_memory_args = {'_c': [], '_by': set()}
        for col in args.get('_c', []):
            name, agg = col.rsplit(_agg_sep, 1) if _agg_sep in col else (col, None)
            for c in name.split(separator):
                in_memory_args['_by'].add(c)
        # Apply all filters while fetching, if skip the filter columns.
        # We'll apply THOSE filters independently
        for key, vals in args.items():
            col = key
            for op in operators:
                if col.endswith(op):
                    col = col[: -len(op)]
                    break
            if _agg_sep in col:
                col = col.rsplit(_agg_sep, 1)[0]
            if col not in in_memory_args['_by']:
                in_memory_args[key] = vals
        url = filter(url, args=in_memory_args, **kwargs)

    # Get unique values for each column
    for col in args.get('_c', []):
        # If ?_c=sales|RANGE, get the range
        name, agg = col.rsplit(_agg_sep, 1) if _agg_sep in col else (col, None)
        # If ?_c=a,b then treat columns a and b as a pair
        cols = name.split(separator)
        # col_args takes _sort, _c and all filters from args
        col_args = {}
        for key, value in args.items():
            # Apply only _sort as a control. Ignore _by, _limit, _offset, etc.
            if key in ['_sort']:
                col_args[key] = value
            # Apply filters. But ignore filters on the columns we're currently processing
            if not key.startswith('_') and key not in cols:
                col_args[key] = value
        if agg:
            # Convert ?_c=col|RANGE to ?_c=col|MIN&_c=col|MAX
            aggs = ['min', 'max'] if agg.lower() == 'range' else [agg]
            # Group by all values, just return the aggregations
            col_args['_by'] = ['']
            col_args['_c'] = [f'{c}{_agg_sep}{a}' for c in cols for a in aggs]
        else:
            col_args['_by'] = cols
            col_args['_c'] = []
            col_args['_limit'] = [limit]
        result[col] = gramex.data.filter(url, args=col_args, **kwargs)
    return result


def alter(
    url: str, table: str, columns: Dict[str, Union[str, dict]] = None, **kwargs: dict
) -> sa.engine.base.Engine:
    '''Create or alter a table with columns specified.

    Examples:
        >>> gramex.data.alter(url, table, columns={
        ...     'id': {'type': 'int', 'primary_key': True, 'autoincrement': True},
        ...     'when': {'type': 'timestamp', 'default': {'function': 'func.now()'}},
        ...     'email': {'nullable': True, 'default': 'none'},
        ...     'age': {'type': 'float', 'nullable': False, 'default': 18},
        ... })

    Parameters:
        url: sqlalchemy URL
        table: table name
        columns: column names, with values as SQL types or type objects
        **kwargs: passed to `sqlalchemy.create_engine()`.

    Returns:
        SQLAlchemy engine

    If the table exists, new columns (if any) are added. Existing columns are **NOT changed**.

    If the table does not exist, the table is created with the specified columns.

    If there are no changes, the function returns quickly (5 ms), since metadata is cached.

    `columns` can be a dict with values as SQL types (e.g. `"INTEGER"` or `"VARCHAR(10)"`):

        >>> gramex.data.alter(url, table, columns={'id': 'INTEGER', 'name': 'VARCHAR(10)'})

    ... or a dict like `{column_name: type, column_name: type, ...}`:

        >>> gramex.data.alter(url, table, columns={
        ...     'id': {'type': 'int', 'primary_key': True, 'autoincrement': True},
        ...     'when': {'type': 'timestamp', 'default': {'function': 'func.now()'}},
        ...     'email': {'nullable': True, 'default': 'none'},
        ...     'age': {'type': 'float', 'nullable': False, 'default': 18},
        ... })

    If the `columns` values are a dict, these keys are allowed:

    - `type` (str): SQL type, e.g. `"VARCHAR(10)"`
    - `default` (str/int/float/bool/function/dict):
        - A scalar like `"none@example.org"` for fixed default values
        - A SQLAlchemy function like `sqlalchemy.func.now()`
        - A dict like `{function: func.now()}` containing a SQLAlchemy functions
    - `nullable` (bool): whether column can have null values, e.g. `False`
    - `primary_key` (bool): whether column is a primary key, e.g. `True`
    - `autoincrement` (bool): whether column automatically increments, e.g. `True`

    `primary_key` and `autoincrement` are used **only** when creating new tables. They do not
    change existing primary keys or autoincrements. This is because
    [SQLite disallows PRIMARY KEY with ALTER](https://stackoverflow.com/a/1120030/100904)
    and AUTO_INCREMENT doesn't work without PRIMARY KEY in MySQL.
    '''
    engine = create_engine(url, **kwargs)
    if columns is None:
        return engine
    # alter is not required for schema-less databases. For now, hard-code engine names
    scheme = urlparse(url).scheme
    if scheme in {'mongodb', 'elasticsearch', 'influxdb'}:
        app_log.info(f'alter() not required for schema-less DB {engine.driver}')
        return engine
    try:
        db_table = get_table(engine, table, extend_existing=True)
    except sa.exc.NoSuchTableError:
        # If the table's not in the DB, create it
        cols = []
        table_kwargs = {"extend_existing": True}
        for name, row in columns.items():
            row = dict({'type': row} if isinstance(row, str) else row, name=name)
            col_type = row.get('type', 'text')
            if isinstance(col_type, str):
                # Use eval() to handle direct types like INTEGER *and* expressions like VARCHAR(3)
                # eval() is safe here since `col_type` is written by app developer
                # B307:eval is safe here since `col_type` is written by app developer
                row['type'] = eval(col_type.upper(), vars(sa.types))  # noqa S307
            row['type_'] = row.pop('type')
            if 'default' in row:
                from inspect import isclass

                default = row.pop('default')
                # default: can be a string like `'sa.func.now()'` or `'func.now()'`
                if isinstance(default, dict):
                    libs = {'sa': sa, 'sqlalchemy': sa, 'func': sa.func}
                    row['server_default'] = build_transform(
                        default, vars={key: None for key in libs}, iter=False
                    )(**libs)
                # default can be an SQLAlchemy function, e.g. sa.func.now()
                elif isclass(default) and issubclass(default, sa.func.Function):
                    row['server_default'] = default
                # default can also be a static value, e.g. `0`
                else:
                    row['server_default'] = str(default)
            if scheme in {'sqlite'} and row.get('autoincrement'):
                table_kwargs['sqlite_autoincrement'] = True
            cols.append(sa.Column(**row))
        sa.Table(table, _METADATA_CACHE[engine], *cols, **table_kwargs).create(engine)
    else:
        quote = engine.dialect.identifier_preparer.quote_identifier
        # If the table's already in the DB, add new columns. We can't change column types
        with engine.connect() as conn, conn.begin():
            for name, row in columns.items():
                if name in db_table.columns:
                    continue
                row = {'type': row} if isinstance(row, str) else row
                col_type = row.get('type', 'text')
                constraints = []
                if 'nullable' in row:
                    constraints.append('' if row['nullable'] else 'NOT NULL')
                if 'default' in row:
                    # repr() converts int, float properly,
                    #   str into 'str' with single quotes (which is the MySQL standard)
                    #   TODO: datetime and other types will fail
                    if isinstance(row['default'], dict) or callable(row['default']):
                        app_log.warning(
                            f'alter(): col {name} cannot change default on existing table'
                        )
                    else:
                        constraints += ['DEFAULT', repr(row['default'])]
                # This syntax works on DB2, MySQL, Oracle, PostgreSQL, SQLite
                conn.execute(
                    f'ALTER TABLE {quote(table)} '
                    f'ADD COLUMN {quote(name)} {col_type} {" ".join(constraints)}'
                )
        # Refresh table metadata after altering
        get_table(engine, table, extend_existing=True)
    return engine


_transform_cache = {}


def _transform_fn(transform, transform_kwargs):
    if callable(transform) and transform_kwargs is not None:
        if transform not in _transform_cache:
            _transform_cache[transform] = lambda v: transform(v, **transform_kwargs)
        return _transform_cache[transform]
    return transform


def _replace(engine, args, *vars, **kwargs):
    escape = _sql_safe if engine == 'sqlalchemy' else _path_safe
    params = {k: v[0] for k, v in args.items() if len(v) > 0 and escape(v[0])}

    def _format(val):
        if isinstance(val, str):
            return val.format(**params)
        if isinstance(val, list):
            return [_format(v) for v in val]
        if isinstance(val, dict):
            return AttrDict([(k, _format(v)) for k, v in val.items()])
        return val

    return _format(list(vars)) + [_format(kwargs)]


def _pop_controls(args):
    '''Filter out data controls: _sort, _limit, _offset, _c (column) and _by from args'''
    return {
        key: args.pop(key) for key in ('_sort', '_limit', '_offset', '_c', '_by') if key in args
    }


def _pop_columns(data, cols, ignored):
    '''Remove columns not in cols'''
    cols = set(cols)
    for col in data.columns:
        if col not in cols:
            ignored.append([col, data[col].tolist()])
    return data[[col for col in cols if col in data.columns]]


def _sql_safe(val):
    '''Return True if val is safe for insertion in an SQL query'''
    if isinstance(val, str):
        return not re.search(r'\s', val)
    elif isinstance(val, (int, float, bool)):
        return True
    return False


def _path_safe(path):
    '''Returns True if path does not try to escape outside a given directory using .. or / etc'''
    # Ignore non-strings. These are generally not meant for paths
    if not isinstance(path, str):
        return True
    return os.path.realpath(os.path.join(_path_safe_root, path)).startswith(_path_safe_root)


# The order of operators is important. ~ is at the end. Otherwise, !~
# or >~ will also be mapped to ~ as an operator
operators = ['!', '>', '>~', '<', '<~', '!~', '~', '!*', '*']


def _filter_col(col, cols):
    '''
    Parses a column name from a list of columns and returns a (col, agg, op)
    tuple.

    - `col` is the name of the column in cols.
    - `agg` is the aggregation operation (SUM, MIN, MAX, etc), else None
    - `op` is the operator ('', !, >, <, etc)

    If the column is invalid, then `col` and `op` are None
    '''
    colset = set(cols)
    # ?col= is returned quickly
    if col in colset:
        return col, None, ''
    # Check if it matches a non-empty operator, like ?col>~=
    for op in operators:
        if col.endswith(op):
            name = col[: -len(op)]
            if name in colset:
                return name, None, op
            # If there's an aggregator, split it out, like ?col|SUM>~=
            elif _agg_sep in name:
                name, agg = name.rsplit(_agg_sep, 1)
                if name in colset:
                    return name, agg, op
    # If no operators match, it might be a pure aggregation, like ?col|SUM=
    if _agg_sep in col:
        name, agg = col.rsplit(_agg_sep, 1)
        if name in colset:
            return name, agg, ''
    # Otherwise we don't know what it is
    return None, None, None


def _argstype(argstype: Dict[str, dict], key: str, default: Any) -> Tuple[Callable, bool]:
    if not isinstance(argstype, dict):
        raise ValueError(f'argstype must be a dict of types, not {argstype!r}')
    argtype = argstype.get(key, {})
    if not isinstance(argtype, dict):
        argtype = {'type': argtype}
    conv = argtype.get('type', default)
    if conv in {'date', 'datetime'}:

        def conv(v):
            return pd.to_datetime(v).to_pydatetime()

    elif isinstance(conv, str):
        conv = build_transform({'function': conv}, iter=False)
    if not callable(conv):
        raise ValueError(f'argstype[{key}][type] must be callable, not {conv!r}')
    return conv, argtype.get('expanding', False)


def _convertor(conv):
    '''Updates a type conversion function.

    - Booleans are converted treating '', '0', 'n', 'no', 'f', 'false' (in any case) as False.
    - Datetimes are converted using pandas to_datetime.
    '''
    # Convert based on Pandas datatype. But for boolean, convert from string as below
    if conv in {np.bool_, bool}:
        conv = lambda v: False if v.lower() in {'', '0', 'n', 'no', 'f', 'false'} else True  # noqa
    elif conv in {datetime}:
        conv = pd.to_datetime
    return conv


def _filter_frame_col(data, key, col, op, vals, conv, meta):
    # Apply type conversion for values
    convert = _convertor(conv)
    vals = tuple(convert(val) for val in vals if val)
    if op not in {'', '!'} and len(vals) == 0:
        meta['ignored'].append((key, vals))
    elif op == '':
        data = data[data[col].isin(vals)] if len(vals) else data[pd.notnull(data[col])]
    elif op == '!':
        data = data[~data[col].isin(vals)] if len(vals) else data[pd.isnull(data[col])]
    elif op == '>':
        data = data[data[col] > min(vals)]
    elif op == '>~':
        data = data[data[col] >= min(vals)]
    elif op == '<':
        data = data[data[col] < max(vals)]
    elif op == '<~':
        data = data[data[col] <= max(vals)]
    # Note: If data[col] contains nans, then the .str.contains() also has NaNs. This raises a
    # ValueError: Cannot mask with non-boolean array containing NA / NaN values.
    # So we use .fillna(False) to convert NaNs to False
    elif op == '!~':
        data = data[~data[col].str.contains('|'.join(vals)).fillna(False)]
    elif op == '~':
        data = data[data[col].str.contains('|'.join(vals)).fillna(False)]
    elif op == '!*':
        data = data[~data[col].str.contains('|'.join(vals), case=False).fillna(False)]
    elif op == '*':
        data = data[data[col].str.contains('|'.join(vals), case=False).fillna(False)]
    meta['filters'].append((col, op, vals))
    return data


def _filter_db_col(query, method, key, col, op, vals, column, conv, meta):
    '''
    - Updates `query` with a method (WHERE/HAVING) that sets '<key> <op> <vals>'
    - `column` is the underlying ColumnElement
    - `conv` is a type conversion function that converts `vals` to the correct type
    - Updates `meta` with the fields used for filtering (or ignored)
    '''
    convert = _convertor(conv)
    vals = tuple(convert(val) for val in vals if val)
    if op not in {'', '!'} and len(vals) == 0:
        meta['ignored'].append((key, vals))
    elif op == '':
        # Test if column is not NULL. != None is NOT the same as is not None
        query = method(column.in_(vals) if len(vals) else column != None)  # noqa
    elif op == '!':
        # Test if column is NULL. == None is NOT the same as is None
        query = method(column.notin_(vals) if len(vals) else column == None)  # noqa
    elif op == '>':
        query = method(column > min(vals))
    elif op == '>~':
        query = method(column >= min(vals))
    elif op == '<':
        query = method(column < max(vals))
    elif op == '<~':
        query = method(column <= max(vals))
    elif op == '!~':
        query = method(column.notlike('%' + '%'.join(vals) + '%'))
    elif op == '~':
        query = method(column.like('%' + '%'.join(vals) + '%'))
    elif op == '!*':
        query = method(column.notilike('%' + '%'.join(vals) + '%'))
    elif op == '*':
        query = method(column.ilike('%' + '%'.join(vals) + '%'))
    meta['filters'].append((col, op, vals))
    return query


def _filter_sort_columns(controls: dict, cols: List[str], meta: dict) -> List[str]:
    '''
    Checks ?_sort=col&_sort=-col. Returns list of (columns to sort by, ascending/not).

    Updates meta['sort'] with columns to sort by,
    and meta['ignored'] with unrecognized sort column names as ('_sort', columns)
    '''
    sorts, ignore_sorts = [], []
    if '_sort' in controls:
        for col in controls['_sort']:
            if col in cols:
                sorts.append((col, True))
            elif col.startswith('-') and col[1:] in cols:
                sorts.append((col[1:], False))
            else:
                ignore_sorts.append(col)
        if len(ignore_sorts) > 0:
            meta['ignored'].append(('_sort', ignore_sorts))
        meta['sort'] = sorts
    return sorts


def _filter_select_columns(controls, cols, meta):
    '''
    Checks ?_c=col&_c=-col. Takes values of ?_c= as col_filter and
    data column names as cols. Returns list of columns to show.

    Updates meta['excluded'] with columns explicitly excluded,
    and meta['ignored'] with unrecognized column names.
    '''
    if '_c' not in controls:
        return cols
    selected_cols, excluded_cols, ignored_cols = [], set(), []
    for col in controls['_c']:
        if col in cols:
            selected_cols.append(col)
        elif col.startswith('-') and col[1:] in cols:
            excluded_cols.add(col[1:])
        else:
            ignored_cols.append(col)
    if len(excluded_cols) > 0 and len(selected_cols) == 0:
        selected_cols = cols
    show_cols = [col for col in selected_cols if col not in excluded_cols]
    meta['excluded'] = list(excluded_cols)
    if ignored_cols:
        meta['ignored'].append(('_c', ignored_cols))
    return show_cols


def _filter_offset_limit(controls, meta):
    offset, limit = 0, None
    if '_offset' in controls:
        try:
            offset = min(int(v) for v in controls['_offset'])
        except ValueError:
            raise ValueError(f'_offset not integer: {controls["_offset"]!r}')
        meta['offset'] = offset
    if '_limit' in controls:
        try:
            limit = min(int(v) for v in controls['_limit'])
        except ValueError:
            raise ValueError(f'_limit not integer: {controls["_limit"]!r}')
        meta['limit'] = limit
    return offset, limit


def _filter_groupby_columns(by, cols, meta):
    '''
    Checks ?_by=col&_by=col for filter().

    - `by`: list of column names to group by
    - `cols`: list of valid column names
    - `meta`: meta['by'] and meta['ignored'] are updated

    Returns a list of columns to group by
    '''
    colset = set(cols)
    for col in by:
        if col in colset and col not in meta['by']:
            meta['by'].append(col)
        else:
            meta['ignored'].append(('_by', col))
    return meta['by']


# If ?by=col|avg is provided, this works in SQL but not in Pandas DataFrames.
# Convert into a DataFrame friendly function
_frame_functions = {
    'avg': 'mean',
    'average': 'mean',
}


def _filter_frame(
    data: pd.DataFrame,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    source: str = 'select',
    id: List[str] = None,
) -> pd.DataFrame:
    '''
    If `source` is `'select'`, returns a DataFrame in which the DataFrame
    `data` is filtered using `args`. Additional controls like _sort, etc are
    in `controls`. Metadata is stored in `meta`.

    If `source` is `'update'`, filters using `args` but only for columns
    mentioned in `id`. Resulting DataFrame is updated with remaining `args`.
    Returns the updated rows.

    If `source` is `'delete'`, filters using `args` but only for columns
    mentioned in `id`. Deletes these rows. Returns the deleted rows.

    Parameters:
        data: dataframe
        meta: dictionary of `filters`, `ignored`, `sort`, `offset`, `limit` params from kwargs
        args: user arguments to filter the data
        source: accepted values - `update`, `delete` for PUT, DELETE methods in FormHandler
        id: list of id specific to data using which values can be updated
    '''
    original_data = data
    cols_for_update = {}
    cols_having = []
    for key, vals in args.items():
        # check if `key` is in the `id` list -- ONLY when data is updated
        if (source == 'update' and key in id) or source in {'select', 'delete'}:
            # Parse column names, ignoring missing / unmatched columns
            col, agg, op = _filter_col(key, data.columns)
            if col is None:
                meta['ignored'].append((key, vals))
                continue
            # Process aggregated columns AFTER filtering, not before (like HAVING clause)
            # e.g. ?sales|SUM=<val> should be applied only after the column is created
            if agg is not None:
                cols_having.append((key, col + _agg_sep + agg, op, vals))
                continue
            # If arg's type is not specified, use the column type
            conv, expanding = _argstype(argstype, key, data[col].dtype.type)
            # Apply filters
            data = _filter_frame_col(data, key, col, op, vals, data[col].dtype.type, meta)
        elif source == 'update':
            # Update values should only contain 1 value. 2nd onwards are ignored
            if key not in data.columns or len(vals) == 0:
                meta['ignored'].append((key, vals))
            else:
                cols_for_update[key] = vals[0]
                if len(vals) > 1:
                    meta['ignored'].append((key, vals[1:]))
        else:
            meta['ignored'].append((key, vals))
    meta['count'] = len(data)
    if source == 'delete':
        original_data.drop(data.index, inplace=True)
        return data
    elif source == 'update':
        conv = {k: v.type for k, v in data.dtypes.items()}
        for key, val in cols_for_update.items():
            original_data.loc[data.index, key] = conv[key](val)
        return data
    else:
        # Apply controls
        if '_by' in controls:
            by = _filter_groupby_columns(controls['_by'], data.columns, meta)
            # If ?_c is not specified, use 'col|sum' for all numeric columns
            # TODO: This does not support ?_c=-<col> to hide a column
            col_list = controls.get('_c')
            if col_list is None:
                col_list = [
                    col + _agg_sep + 'sum'
                    for col in data.columns
                    if pd.api.types.is_numeric_dtype(data[col])
                ]
            agg_cols = []
            agg_dict = AttrDict()
            for key in col_list:
                col, agg, val = _filter_col(key, data.columns)
                if agg is not None:
                    # Convert aggregation into a Pandas GroupBy agg function
                    agg = agg.lower()
                    agg = _frame_functions.get(agg, agg)
                    agg_cols.append(key)
                    if col in agg_dict:
                        agg_dict[col].append(agg)
                    else:
                        agg_dict[col] = [agg]
            if len(by) > 0:
                if not agg_cols:
                    # If no aggregation columns exist, just show groupby columns.
                    data = data.groupby(by).agg('size').reset_index()
                    data = data.iloc[:, : len(by)]
                else:
                    data = data.groupby(by).agg(agg_dict)
                    data.columns = agg_cols
                    data = data.reset_index()
                # Apply HAVING operators
                for key, col, op, vals in cols_having:
                    # If arg's type is not specified, use the column type
                    conv, expanding = _argstype(argstype, key, data[col].dtype.type)
                    data = _filter_frame_col(data, key, col, op, vals, conv, meta)
            else:
                row = [data[col].agg(op) for col, ops in agg_dict.items() for op in ops]
                data = pd.DataFrame([row], columns=agg_cols)
        elif '_c' in controls:
            show_cols = _filter_select_columns(controls, data.columns, meta)
            data = data[show_cols]
        sorts = _filter_sort_columns(controls, data.columns, meta)
        if sorts:
            data = data.sort_values(by=[c[0] for c in sorts], ascending=[c[1] for c in sorts])
        offset, limit = _filter_offset_limit(controls, meta)
        if offset is not None:
            data = data.iloc[offset:]
        if limit is not None:
            data = data.iloc[:limit]
        return data


def _filter_db(
    engine: str,
    table: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    source: str = 'select',
    id: List[str] = None,
):
    '''
    Parameters:
        engine: constructed sqlalchemy string
        table: table name in the mentioned database
        meta: dictionary of `filters`, `ignored`, `sort`, `offset`, `limit` params from kwargs
        controls: dictionary of `_sort`, `_c`, `_offset`, `_limit` params
        source: accepted values - `update`, `delete` for PUT, DELETE methods in FormHandler
        args: dictionary of user arguments to filter the data
        argstype: optional dict that specifies `args` type and behavior.
        id: list of keys specific to data using which values can be updated
    '''
    table = get_table(engine, table)
    cols = table.columns
    colslist = cols.keys()

    if source == 'delete':
        query = sa.delete(table)
    elif source == 'update':
        query = sa.update(table)
    else:
        query = sa.select([table])
    cols_for_update = {}
    cols_having = []
    for key, vals in args.items():
        # check if `key` is in the `id` list -- ONLY when data is updated
        if (source == 'update' and key in id) or source in {'select', 'delete'}:
            # Parse column names, ignoring missing / unmatched columns
            col, agg, op = _filter_col(key, colslist)
            if col is None:
                meta['ignored'].append((key, vals))
                continue
            # Process aggregated columns AFTER filtering, not before (like HAVING clause)
            # e.g. ?sales|SUM=<val> should be applied only after the column is created
            if agg is not None:
                cols_having.append((key, col + _agg_sep + agg, op, vals))
                continue
            # If arg's type is not specified, use the column type
            conv, expanding = _argstype(argstype, key, cols[col].type.python_type)
            # Apply filters
            query = _filter_db_col(query, query.where, key, col, op, vals, cols[col], conv, meta)
        elif source == 'update':
            # Update values should only contain 1 value. 2nd onwards are ignored
            if key not in cols or len(vals) == 0:
                meta['ignored'].append((key, vals))
            else:
                cols_for_update[key] = vals[0]
                if len(vals) > 1:
                    meta['ignored'].append((key, vals[1:]))
        else:
            meta['ignored'].append((key, vals))
    if source == 'delete':
        res = engine.execute(query)
        return res.rowcount
    elif source == 'update':
        query = query.values(cols_for_update)
        res = engine.execute(query)
        return res.rowcount
    else:
        # Apply controls
        if '_by' in controls:
            by = _filter_groupby_columns(controls['_by'], colslist, meta)
            query = query.group_by(*by)
            # If ?_c is not specified, use 'col|sum' for all numeric columns
            # TODO: This does not support ?_c=-<col> to hide a column
            col_list = controls.get('_c')
            if col_list is None:
                col_list = [
                    col + _agg_sep + 'sum'
                    for col, column in cols.items()
                    if column.type.python_type.__name__ in _numeric_types
                ]
            agg_cols = AttrDict([(col, cols[col]) for col in by])  # {label: ColumnElement}
            typ = {}  # {label: python type}
            for key in col_list:
                col, agg, val = _filter_col(key, colslist)
                if agg is not None:
                    # Convert aggregation into SQLAlchemy query
                    agg = agg.lower()
                    typ[key] = _agg_type.get(agg, cols[col].type.python_type)
                    agg_func = getattr(sa.sql.expression.func, agg)
                    agg_cols[key] = agg_func(cols[col]).label(key)
            if not agg_cols:
                return pd.DataFrame()
            # SQLAlchemy 1.4+ only accepts positional arguments for .with_only_columns()
            if version.parse(sa.__version__) >= version.parse('1.4'):
                query = query.with_only_columns(*agg_cols.values())
            # SQLAlchemy 1.3- only accepts a list for .with_only_columns()
            else:
                query = query.with_only_columns(agg_cols.values())
            # Apply HAVING operators
            for key, col, op, vals in cols_having:
                query = _filter_db_col(
                    query, query.having, key, col, op, vals, agg_cols[col], typ[col], meta
                )
        elif '_c' in controls:
            show_cols = _filter_select_columns(controls, colslist, meta)
            query = query.with_only_columns([cols[col] for col in show_cols])
            if len(show_cols) == 0:
                return pd.DataFrame()
        # SQLAlchemy 1.4+ deprecated SelectBase.columns in favor of SelectBase.selected_columns
        try:
            selected_columns = query.selected_columns
        except AttributeError:
            selected_columns = query.columns
        sortable_columns = colslist + selected_columns.keys()
        sorts = _filter_sort_columns(controls, sortable_columns, meta)
        for col, asc in sorts:
            orderby = sa.asc if asc else sa.desc
            query = query.order_by(orderby(col))
        offset, limit = _filter_offset_limit(controls, meta)
        # MSSQL does not support OFFSET without ORDER BY. So sort by first column
        if (
            'mssql' in engine.url.drivername
            and (offset is not None or limit is not None)
            and not sorts
        ):
            query = query.order_by(sortable_columns[0])
        if offset is not None:
            query = query.offset(offset)
        if limit is not None:
            query = query.limit(limit)
        return pd.read_sql(query, engine)


_VEGA_SCRIPT = os.path.join(_FOLDER, 'download.vega.js')


# MongoDB Operations
# ----------------------------------------

_mongodb_op_map = {'<': '$lt', '<~': '$lte', '>': '$gt', '>~': '$gte', '': '$in', '!': '$nin'}


def _mongodb_query(args, table, id: List[str] = [], **kwargs):
    # Convert a query like x>=3&x>=4&x>=5 into
    # {"$or": [{x: {$gt: 3}}, {x: {$gt: 4}}, {x: $gt: 5}]}
    row = table.find_one()
    if not row:
        return {}
    conditions = []
    for key, vals in args.items():
        if len(id) and key not in id:
            continue
        col_names = list(row)
        col, agg, op = _filter_col(key, col_names)
        if not col:
            continue
        add = lambda v: conditions.append({col: v})  # noqa
        convert = _convertor(type(row[col])) if col in row else lambda v: v
        if op in {'', '!'}:
            add({_mongodb_op_map[op]: [convert(val) for val in vals if val]})
            if any(not val for val in vals):
                add({'$eq' if op == '!' else '$ne': None})
        elif op in {'>', '>~', '<', '<~'}:
            vals = [convert(val) for val in vals if val]
            if op == '>':
                add({_mongodb_op_map[op]: max(vals)})
            elif op == '<':
                add({_mongodb_op_map[op]: min(vals)})
            elif op == '>~':
                add({_mongodb_op_map[op]: max(vals)})
            elif op == '<~':
                add({_mongodb_op_map[op]: min(vals)})
        elif op == '!~':
            add({"$not": {"$regex": '|'.join(vals)}})
        elif op == '~':
            add({"$regex": '|'.join(vals)})
        elif op == '!*':
            add({"$not": {"$regex": '|'.join(vals), "$options": 'i'}})
        elif op == '*':
            add({"$regex": '|'.join(vals), "$options": 'i'})
        elif col and op in _mongodb_op_map:
            add({_mongodb_op_map[op]: convert(val)} for val in vals)
        # TODO: Handle agg
        # TODO: add meta['ignored']
    return {'$and': conditions} if len(conditions) > 1 else conditions[0] if conditions else {}


def _mongodb_collection(url: str, database: str, collection: str, **kwargs):
    import pymongo

    # Support all MongoDB client arguments
    # https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html
    mongo_kwargs = {
        'host',
        'port',
        'document_class',
        'tz_aware',
        'connect',
        'type_registry',
        # Other optional parameters can be passed as keyword arguments
        'directConnection',
        'maxPoolSize',
        'minPoolSize',
        'maxIdleTimeMS',
        'maxConnecting',
        'socketTimeoutMS',
        'connectTimeoutMS',
        'server_selector',
        'serverSelectionTimeoutMS',
        'waitQueueTimeoutMS',
        'heartbeatFrequencyMS',
        'appname',
        'driver',
        'event_listeners',
        'retryWrites',
        'retryReads',
        'compressors',
        'zlibCompressionLevel',
        'uuidRepresentation',
        'unicode_decode_error_handler',
        'srvServiceName',
        # Write Concern options
        'w',
        'wTimeoutMS',
        'journal',
        'fsync',
        # Read Concern options
        'readConcernLevel',
        # Replica set keyword arguments
        'replicaSet',
        # Read Preference
        'readPreference',
        'readPreferenceTags',
        'maxStalenessSeconds',
        # Authentication
        'username',
        'password',
        'authSource',
        'authMechanism',
        'authMechanismProperties',
        # TLS/SSL configuration
        'tls',
        'tlsInsecure',
        'tlsAllowInvalidCertificates',
        'tlsAllowInvalidHostnames',
        'tlsCAFile',
        'tlsCertificateKeyFile',
        'tlsCRLFile',
        'tlsCertificateKeyFilePassword',
        'tlsDisableOCSPEndpointCheck',
        'ssl',
        # Client side encryption options
        'auto_encryption_opts',
        # Versioned API options
        'server_api',
    }
    create_kwargs = {key: val for key, val in kwargs.items() if key in mongo_kwargs}
    client = create_engine(url, create=pymongo.MongoClient, **create_kwargs)
    db = client[database]
    return db[collection]


def _mongodb_json(obj):
    # Parse val in keys ending with . as JSON ({"key.": val}), but retain other keys
    result = {}
    for key, val in obj.items():
        if key.endswith('.'):
            result[key[:-1]] = json.loads(val)
        else:
            result[key] = val
    return result


def _filter_mongodb(
    url: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for collection?
    ext: str = None,
    database: str = None,
    collection: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    # TODO: Document function and usage
    table = _mongodb_collection(url, database, collection, **kwargs)
    # TODO: If data is missing, identify columns using columns:
    row = table.find_one(query)
    if not row:
        return pd.DataFrame()
    cols = list(row)
    query = dict(query) if query else _mongodb_query(args, table)

    show_cols = _filter_select_columns(controls, cols, meta)
    if show_cols:
        cursor = table.find(query, show_cols)
    else:
        cursor = table.find(query)
    sorts = _filter_sort_columns(controls, cols, meta)
    if sorts:
        cursor = cursor.sort([(key, +1 if val else -1) for key, val in sorts])
    offset, limit = _filter_offset_limit(controls, meta)
    if offset is not None:
        cursor = cursor.skip(offset)
    if limit is not None:
        cursor = cursor.limit(limit)

    data = pd.DataFrame(list(cursor))
    # Convert Object IDs into strings to allow JSON conversion
    if len(data) > 0:
        import bson

        for col, val in data.iloc[0].items():
            if type(val) in {bson.objectid.ObjectId}:
                data[col] = data[col].map(str)

    return data


def _delete_mongodb(
    url: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for collection?
    ext: str = None,
    database: str = None,
    collection: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    table = _mongodb_collection(url, database, collection, **kwargs)
    query = _mongodb_query(args, table)
    result = table.delete_many(query)
    return result.deleted_count


def _update_mongodb(
    url: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for collection?
    ext: str = None,
    database: str = None,
    collection: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    table = _mongodb_collection(url, database, collection, **kwargs)
    query = _mongodb_query(args, table, id=id)
    row = table.find_one(query)
    if not row:
        return 0

    values = {key: val[-1] for key, val in dict(args).items() if key not in id}
    for key in values.keys():
        convert = _convertor(type(row[key])) if key in row else lambda v: v
        values[key] = convert(values[key])

    result = table.update_many(query, {'$set': _mongodb_json(values)})
    return result.modified_count


def _insert_mongodb(
    url: str,
    meta: dict,
    controls: dict,
    rows: list,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for collection?
    ext: str = None,
    database: str = None,
    collection: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    table = _mongodb_collection(url, database, collection, **kwargs)
    result = table.insert_many([_mongodb_json(row) for row in rows.to_dict(orient='records')])
    meta['inserted'] = [{'id': str(id) for id in result.inserted_ids}]  # noqa B035
    return len(result.inserted_ids)


# InfluxDB Operations
# ----------------------------------------


def _get_influxdb_schema(client, bucket: str):
    imports = 'import "influxdata/influxdb/schema"\n'
    meas = client.query_api().query(imports + f'schema.measurements(bucket: "{bucket}")')[0]
    tags = client.query_api().query(imports + f'schema.tagKeys(bucket: "{bucket}")')[0]
    tags = [r.get_value() for r in tags.records]
    tags = [r for r in tags if not r.startswith("_")]
    fields = client.query_api().query(imports + f'schema.fieldKeys(bucket: "{bucket}")')[0]
    return {
        "_measurement": [r.get_value() for r in meas.records],
        "_tags": tags,
        "_fields": [r.get_value() for r in fields.records],
    }


_influxdb_op_map = {"<~": "<=", ">~": ">=", "!": "!=", "": "=="}


def _influxdb_offset_limit(controls):
    offset = controls.pop("_offset", ["-30d"])[0]
    limit = controls.pop("_limit", [False])[0]
    offset = f"start: {offset}"
    if isinstance(limit, str) and not limit.isdigit():
        offset += f", stop: {limit}"
        limit = False

    return offset, limit


def _filter_influxdb(
    url: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for org/bucket?
    ext: str = None,
    org: str = None,
    bucket: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    with _influxdb_client(url, org=org, **kwargs) as db:
        schema = _get_influxdb_schema(db, bucket)
        cols = schema["_fields"] + schema["_tags"] + schema["_measurement"]
        q = db.query_api()
        offset, limit = _influxdb_offset_limit(controls)
        query = f'from(bucket: "{bucket}")|>range({offset})\n'

        filters = []
        wheres = []
        to_drop = []
        for col in controls.pop("_c", []):
            if col.startswith("-"):
                if col[1:] in schema["_fields"]:
                    wheres.append(f'r._field != "{col[1:]}"')
                else:
                    to_drop.append(col[1:])
            elif col in schema["_fields"]:
                col, agg, op = _filter_col(col, cols)
                op = _influxdb_op_map.get(op, op)
                wheres.append(f'r._field {op} "{col}"')
        if len(wheres):
            wheres = " or ".join(wheres)
            filters.append(f"|> filter(fn: (r) => {wheres})")
        for key, vals in args.items():
            col, agg, op = _filter_col(key, cols)
            op = _influxdb_op_map.get(op, op)
            if col in schema["_fields"]:
                where = " or ".join([f'r._field == "{col}" and r._value {op} {v}' for v in vals])
            else:
                where = " or ".join([f'r["{col}"] {op} "{v}"' for v in vals])
            filters.append(f"|> filter(fn: (r) => {where})")
        query += "\n".join(filters)
        if to_drop:
            to_drop = ",".join([f'"{k}"' for k in to_drop])
            query += f"\n|> drop(columns: [{to_drop}])"

        if limit:
            query += f"|> tail(n: {limit})\n"

        app_log.debug("Running InfluxDB query: \n" + query)

        df = q.query_data_frame(query)
    return df.drop(["result", "table"], axis=1, errors="ignore")


def _delete_influxdb(
    url: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for org/bucket?
    ext: str = None,
    org: str = None,
    bucket: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    with _influxdb_client(url, org=org, **kwargs) as db:
        schema = _get_influxdb_schema(db, bucket)
        start = args.pop('_time>', ['0'])[0]
        stop = args.pop('_time<', ['99999999999'])[0]
        measurement = args.pop('_measurement', schema['_measurement'])[0]
        predicate = f'_measurement="{measurement}"'
        tags = [(k, v[0]) for k, v in args.items() if k in schema['_tags']]
        if tags:
            tag_predicate = " OR ".join([f'{k}="{v}"' for k, v in tags])
            predicate += ' AND ' + f'({tag_predicate})'
        db.delete_api().delete(start, stop, predicate, bucket, org)
    # InfluxDB Python client doesn't return anything on DELETE, we return a mock number here.
    return 0


def _influxdb_client(
    url: str,
    token: str,
    org: str,
    debug: Any = None,
    timeout: int = 60_000,
    enable_gzip: bool = False,
    default_tags: dict = None,
    **kwargs,
):
    from influxdb_client import InfluxDBClient

    url = re.sub(r"^influxdb:", "", url)
    return InfluxDBClient(
        url,
        token,
        org=org,
        debug=debug,
        enable_gzip=enable_gzip,
        default_tags=default_tags,
        timeout=timeout,
    )


def _timestamp_df(df, index_col="_time"):
    """Add timestamp index by parsing `index_col` (default to now)."""
    now = datetime.now().astimezone()
    if index_col not in df:
        df[index_col] = [now] * len(df)
    else:
        df[index_col] = pd.to_datetime(df[index_col], errors="coerce").fillna(now)
    return df.set_index(index_col)


def _get_ts_points(df, measurement, tags):
    from influxdb_client import Point

    points = []
    tags = df[tags]
    fields = df.drop(tags, axis=1)
    for (t, field), (_, tag) in zip(fields.astype(float).iterrows(), tags.iterrows()):
        p = Point(measurement).time(t)
        [p.tag(t, tval) for t, tval in tag.to_dict().items()]
        [p.field(f, fval) for f, fval in field.to_dict().items()]
        points.append(p)
    return points


def _insert_influxdb(
    url: str,
    meta: dict,
    controls: dict,
    rows: list,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,  # TODO: Should table be an alias for org/bucket?
    ext: str = None,
    org: str = None,
    bucket: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    measurement = rows.pop("measurement").unique()[0]
    tags = rows.pop("tags").dropna().drop_duplicates().tolist() if "tags" in rows else []
    # Ensure that the index is timestamped
    rows = _timestamp_df(rows)

    # Ensure that all columns except tags are floats
    field_columns = [c for c in rows if c not in tags]
    rows[field_columns] = rows[field_columns].astype(float)
    from influxdb_client.client.write_api import ASYNCHRONOUS, WriteOptions

    with _influxdb_client(url, **kwargs) as db, db.write_api(
        write_options=WriteOptions(ASYNCHRONOUS, batch_size=50000, flush_interval=10000)
    ) as client:
        client.write(
            bucket=bucket,
            org=db.org,
            record=rows,
            data_frame_measurement_name=measurement,
            data_frame_tag_columns=tags,
        )
    meta["inserted"] = [{"id": ix} for ix, _ in rows.iterrows()]
    return len(rows)


# ServiceNow Operations
# ----------------------------------------


def _filter_servicenow(
    url: str,
    meta: dict,
    controls: dict,
    args: dict,
    argstype: Dict[str, dict] = {},
    id: List[str] = None,
    table: str = None,
    ext: str = None,
    database: str = None,
    collection: str = None,
    columns: Dict[str, Union[str, dict]] = None,
    query: str = None,
    queryfile: str = None,
    **kwargs,
):
    import pysnow
    from gramex.config import locate

    urlinfo = urlparse(url)
    c = pysnow.Client(instance=urlinfo.hostname, user=urlinfo.username, password=urlinfo.password)

    config = gramex.cache.open('servicenow.yaml', 'config', rel=True)
    if table is None:
        table = c.resource(api_path=urlinfo.path)

    # Identify column types. Take the default ServiceNow column types. Override based on user
    table_columns = json.loads(json.dumps(config.columns[urlinfo.path]))
    for col, colinfo in (columns or {}).items():
        table_columns.setdefault(col, {}).update(
            colinfo if isinstance(colinfo, dict) else {'type': colinfo}
        )
    for colinfo in table_columns.values():
        colinfo['_convert'] = _convertor(locate(colinfo['type'], modules=['datetime']))
    col_names = table_columns.keys()

    fields = []
    if not query:
        offset, limit = _filter_offset_limit(controls, meta)
        query = pysnow.QueryBuilder()
        query_initiated = False

        for key, value in args.items():
            col, agg, op = _filter_col(key, col_names)
            value = [table_columns[col]['_convert'](val) for val in value]
            if query_initiated:
                query.AND()
            query_initiated = True
            if op == '':
                query.field(col).equals(value)
            elif op == '!':
                query.field(col).not_equals(value)
            elif op == '>':
                query.field(col).greater_than(min(value))
            elif op == '>~':
                query.field(col).greater_than_or_equal(min(value))
            elif op == '<':
                query.field(col).less_than(max(value))
            elif op == '<~':
                query.field(col).less_than_or_equal(max(value))
            elif op == '!~':
                query.field(col).not_contains(value[0])
            elif op == '~':
                query.field(col).contains(value[0])
            # TODO: Handle case insensitive search via * and !* operators for ServiceNow
            else:
                raise ValueError(f'Unknown ServiceNow operator: {op}')

    offset, limit = _filter_offset_limit(controls, meta)
    sorts = _filter_sort_columns(controls, col_names, meta)
    for col, asc in sorts:
        if query_initiated:
            query.AND()
        query_initiated = True
        if asc:
            query.field(col).order_ascending()
        else:
            query.field(col).order_descending()
    fields.extend(_filter_select_columns(controls, col_names, meta))

    if not len(query._query):
        query = {}

    response = table.get(query=query, limit=limit, offset=offset, fields=fields)
    return pd.DataFrame(response.all())


plugins["mongodb"] = {
    "filter": _filter_mongodb,
    "delete": _delete_mongodb,
    "insert": _insert_mongodb,
    "update": _update_mongodb,
}
plugins["influxdb"] = {
    "filter": _filter_influxdb,
    "delete": _delete_influxdb,
    "insert": _insert_influxdb,
    "update": _insert_influxdb,
}
plugins["servicenow"] = {"filter": _filter_servicenow}



import contextlib
import os
import six
import json
import time
import logging
import mimetypes
import traceback
import tornado.gen
import gramex
import gramex.cache
from typing import Union, Optional, List, Dict, Any
from binascii import b2a_base64
from fnmatch import fnmatch
from http.cookies import Morsel
from orderedattrdict import AttrDict
from urllib.parse import urlsplit, urlunsplit, parse_qsl, urljoin, urlencode
from tornado.web import RequestHandler, HTTPError, MissingArgumentError, decode_signed_value
from tornado.websocket import WebSocketHandler
from gramex import conf, __version__
from gramex.config import merge, objectpath, app_log
from gramex.transforms import build_transform, build_log_info, handler_expr, time_key
from gramex.transforms.template import CacheLoader
from gramex.http import UNAUTHORIZED, FORBIDDEN, BAD_REQUEST, METHOD_NOT_ALLOWED, TOO_MANY_REQUESTS
from gramex.cache import get_store

# We don't use these, but these stores used to be defined here. Programs may import these
from gramex.cache import KeyStore, JSONStore, HDF5Store, SQLiteStore, RedisStore  # noqa

server_header = f'Gramex/{__version__}'
_store_cache = {}

# Python 3.8+ supports SameSite cookie attribute. Monkey-patch it for Python 3.7
# https://stackoverflow.com/a/50813092/100904
Morsel._reserved.setdefault('samesite', 'SameSite')


class BaseMixin:
    '''Common utilities for all handlers. This is used by
    [BaseHandler][gramex.handlers.BaseHandler] and
    [BaseWebSocketHandler][gramex.handlers.BaseWebSocketHandler].
    '''

    @classmethod
    def setup(
        cls,
        transform={},
        redirect={},
        methods=None,
        auth: Union[None, bool, dict] = None,
        log=None,
        set_xsrf=None,
        error=None,
        xsrf_cookies=None,
        cors: Union[None, bool, dict] = None,
        ratelimit: Optional[dict] = None,
        validate: Optional[Union[dict, list, str]] = None,
        # If you add any explicit kwargs here, add them to special_keys too.
        **kwargs,
    ):
        '''
        One-time setup for all request handlers. This is called only when
        gramex.yaml is parsed / changed.
        '''
        cls._on_init_methods = []
        cls._on_finish_methods = []
        cls._set_xsrf = set_xsrf

        cls.kwargs = cls.conf.get('kwargs', AttrDict())

        cls.setup_transform(transform)
        cls.setup_redirect(redirect)
        # Note: call setup_session before setup_auth to ensure that .session is available.
        # This also ensures we override_user before auth
        cls.setup_session(conf.app.get('session'))
        cls.setup_ratelimit(ratelimit, conf.app.get('ratelimit'))
        cls.setup_auth(auth)
        cls.setup_error(error)
        cls.setup_xsrf(xsrf_cookies)
        cls.setup_log()
        cls.setup_httpmethods(methods)
        cls.setup_cors(cors, auth=auth)
        cls.setup_validate(validate)

        # app.settings.debug enables debugging exceptions using pdb
        if conf.app.settings.get('debug', False):
            cls.log_exception = cls.debug_exception

    # A list of special keys handled by BaseHandler. Can be extended by other classes.
    special_keys = [
        'validate',
        'transform',
        'redirect',
        'methods',
        'auth',
        'log',
        'set_xsrf',
        'error',
        'xsrf_cookies',
        'cors',
        'headers',
        'ratelimit',
    ]

    @classmethod
    def clear_special_keys(cls, kwargs, *args):
        '''
        Remove keys handled by BaseHandler that may interfere with setup().
        This should be called explicitly in setup() where required.
        '''
        for special_key in cls.special_keys:
            kwargs.pop(special_key, None)
        for special_key in args:
            kwargs.pop(special_key, None)
        return kwargs

    @classmethod
    def get_list(
        cls, val: Union[list, tuple, str], key: str = '', eg: str = '', caps: bool = True
    ) -> set:
        '''Split comma-separated values into a set.

        Process kwargs that can be a comma-separated string or a list,
        like BaseMixin's `methods:`, `cors.origins`, `cors.methods`, `cors.headers`,
        `ratelimit.keys`, etc.

        Examples:
            >>> get_list('GET, PUT') == {'GET', 'PUT'}
            >>> get_list(['GET', ' ,get '], caps=True) == {'GET'}
            >>> get_list([' GET ,  PUT', ' ,POST, ']) == {'GET', 'PUT', 'POST'}

        Parameters:
            val: Input to split. If val is not str/list/tuple, raise `ValueError`
            key: `url:` key to display in error message
            eg: Example values to display in error message
            caps: True to convert values to uppercase

        Returns:
            Unique comma-separated values
        '''
        if isinstance(val, (list, tuple)):
            val = ' '.join(val)
        elif not val:
            val = ''
        if not isinstance(val, str):
            err = f'url:{cls.name}.{key}: {val!r} not a string/list'
            err = err + f', e.g. {eg}' if eg else err
            raise ValueError(err)
        if caps:
            val = val.upper()
        # Return de-duplicated list. Avoid set(), use dict to preserve order
        return list(dict.fromkeys(val.replace(',', ' ').split()))

    @classmethod
    def setup_httpmethods(cls, methods: Union[list, tuple, str]):
        methods = cls.get_list(methods, key='methods', eg='[GET, POST]', caps=True)
        if methods:
            cls._http_methods = methods
            cls._on_init_methods.append(cls.check_http_method)

    def check_http_method(self):
        '''If method: [...] is specified, reject all methods not in the allowed methods set'''
        if self.request.method not in self._http_methods:
            raise HTTPError(
                METHOD_NOT_ALLOWED,
                f'{self.name}: method {self.request.method} '
                + f'not in allowed methods {self._http_methods}',
            )

    @classmethod
    def setup_cors(cls, cors: Union[None, bool, dict], auth):
        if cors is None:
            return
        if cors is True:
            cors = {
                'auth': bool(auth),
                'methods': '*',
                'headers': '*',
                'origins': '*',
            }
        if not isinstance(cors, dict):
            app_log.error(f'url:{cls.name}.cors is not a dict/True')
            return
        # Set default CORS values as a set
        for key in ('origins', 'methods', 'headers'):
            cors[key] = cls.get_list(cors.get(key, '*'), f'cors.{key}', '"*"', caps=False)
        cors.setdefault('auth', False)
        cls._on_init_methods.append(cls.check_cors)
        cls.options = cls._cors_options
        cls._cors = cors

    def check_cors(self):
        '''
        For simple CORS requests, send Access-Control-Allow-Origin: <origin>.
        If request needs credentials, allow it.'''
        origin, cred = self.cors_origin()
        if origin:
            self.set_header('Access-Control-Allow-Origin', origin)
            if cred:
                self.set_header('Access-Control-Allow-Credentials', 'true')

    def cors_origin(self):
        '''
        Returns the origin to set in Access-Control-Allow-Origin header.
        '''
        # If CORS is not enabled, it fails
        if not self._cors:
            return None, False
        # Assume credentials are passed if handler requires Auth or Cookie is passed
        cred = self._cors['auth'] or self.request.headers.get('Cookie')
        # If origin: *, then allow all origins
        origin = self.request.headers.get('Origin', '').lower()
        if self._cors['origins'] == set('*'):
            return (origin if cred else '*', cred)
        # If it matches any of the wildcards, return specific origin
        for pattern in self._cors['origins']:
            if fnmatch(origin, pattern.lower()):
                return origin, cred
        # If none of the patterns match, it fails
        return None, cred

    def _cors_options(self, *args, **kwargs):
        # Check if origin is in cors.origin
        origin, cred = self.cors_origin()
        if not origin:
            origin = self.request.headers.get('Origin', '')
            raise HTTPError(
                BAD_REQUEST,
                f'url:{self.name}: CORS origin {origin} not in {self._cors["origins"]}',
            )

        # Check if method is in cors.methods
        method = self.request.headers.get('Access-Control-Request-Method', '').upper()
        for pattern in self._cors['methods']:
            if fnmatch(method, pattern.upper()):
                break
        else:
            raise HTTPError(
                BAD_REQUEST,
                f'url:{self.name}: CORS method {method} not in {self._cors["methods"]}',
            )

        # Check if headers is in cors.headers
        headers = self.request.headers.get('Access-Control-Request-Headers', '')
        headers = self.get_list(headers, 'headers', '', caps=False)
        allowed_headers = {h.lower() for h in self._cors['headers']}
        diff = set()
        if '*' not in allowed_headers:
            for header in headers:
                if header.lower() not in allowed_headers:
                    diff.add(header)
        if diff:
            raise HTTPError(
                BAD_REQUEST,
                f'url:{self.name}: CORS headers {diff} not in {self._cors["headers"]}',
            )

        # If it succeeds, set relevant headers
        self.set_header('Access-Control-Allow-Origin', origin)
        methods = (
            self._all_methods if '*' in self._cors['methods'] else ', '.join(self._cors['methods'])
        )
        self.set_header('Access-Control-Allow-Methods', methods)
        headers = set(headers) | set(self._cors['headers'])
        if '*' in headers:
            headers.remove('*')
            headers.update(self._all_headers)
        self.set_header('Access-Control-Allow-Headers', ', '.join(headers))
        # TODO: Access-Control-Max-Age
        # TODO: Access-Control-Expose-Headers

    _all_methods = 'GET, HEAD, POST, PUT, DELETE, OPTIONS, PATCH, CONNECT, TRACE'
    _all_headers = [
        'Accept',
        'Cache-Control',
        'Content-Type',
        'If-None-Match',
        'Origin',
        'Pragma',
        'Upgrade-Insecure-Requests',
        'X-Requested-With',
    ]

    @classmethod
    def setup_default_kwargs(cls):
        '''
        Use default config from handlers.<Class>.* and handlers.BaseHandler.
        Called directly by gramex.services.url().
        NOTE: This updates the kwargs for setup() -- so it must be called BEFORE setup()
        and can't be merged into setup() or called from setup().
        '''
        c = cls.conf.kwargs
        merge(c, objectpath(conf, 'handlers.' + cls.conf.handler, {}), mode='setdefault')
        merge(c, objectpath(conf, 'handlers.BaseHandler', {}), mode='setdefault')

    @classmethod
    def setup_transform(cls, transform):
        cls.transform = {}
        for pattern, trans in transform.items():
            cls.transform[pattern] = {
                'function': build_transform(
                    trans, vars={'content': None, 'handler': None}, filename=f'url:{cls.name}'
                ),
                'headers': trans.get('headers', {}),
                'encoding': trans.get('encoding'),
            }

    @staticmethod
    def _purge_keys(data):
        '''
        Returns session keys to be deleted. These are either None values or
        those with expired keys based on _t.
        setup_session makes the session store call this method.
        Until v1.20 (31 Jul 2017) no _t keys were set.
        From v1.23 (31 Oct 2017) these are cleared.
        '''
        now = time.time()
        week = 7 * 24 * 60 * 60
        keys = []
        # When using sqlitedict, fetching keys may fail if DB is locked. Try later
        try:
            items = list(data.items())
        except Exception:
            items = []
        for key, val in items:
            # Purge already cleared / removed sessions
            if val is None:
                keys.append(key)
            elif isinstance(val, dict):
                # If the session has expired, remove it
                expired = val.get('_t', 0) < now
                # If the session is inactive, remove it after a week.
                # If we remove immediately, then we may lose WIP sessions.
                # For example, people who opened a login page where _next_url was set
                inactive = '_i' in val and '_l' in val and val['_i'] + val['_l'] < now - week
                if expired or inactive:
                    keys.append(key)
            else:
                app_log.warning(f'Store key: {key} has value type {type(val)} (not dict)')
        return keys

    @classmethod
    def _get_store(cls, conf):
        key = store_type, store_path = conf.get('type'), conf.get('path')
        if key not in _store_cache:
            _store_cache[key] = get_store(
                type=store_type,
                path=store_path,
                flush=conf.get('flush'),
                purge=conf.get('purge'),
                purge_keys=cls._purge_keys,
            )
        return _store_cache[key]

    @classmethod
    def setup_session(cls, session_conf):
        '''handler.session returns the session object. It is saved on finish.'''
        if session_conf is None:
            return
        cls._session_store = cls._get_store(session_conf)
        cls.session = property(cls.get_session)
        cls._session_expiry = session_conf.get('expiry')
        cls._session_cookie_id = session_conf.get('cookie', 'sid')
        cls._session_cookie = {
            key: session_conf[key]
            for key in ('httponly', 'secure', 'samesite', 'domain')
            if key in session_conf
        }
        # Note: We cannot use path: to specify the Cookie path attribute.
        # session.path is used for the session (JSONStore) file location.
        # So use cookiepath: instead.
        if 'cookiepath' in session_conf:
            cls._session_cookie['path'] = session_conf['cookiepath']
        cls._on_init_methods.append(cls.override_user)
        cls._on_finish_methods.append(cls.set_last_visited)
        # Ensure that session is saved AFTER we set last visited
        cls._on_finish_methods.append(cls.save_session)

    @classmethod
    def setup_ratelimit(
        cls, ratelimit: Union[Dict, List[Dict], None], ratelimit_app_conf: Union[dict, None]
    ):
        '''Initialize rate limiting checks'''
        if ratelimit is None:
            return
        if not ratelimit_app_conf:
            raise ValueError(f"url:{cls.name}.ratelimit: no app.ratelimit defined")

        # All ratelimit related info is stored in self._ratelimit
        cls._ratelimit = []
        cls._on_init_methods.append(cls.check_ratelimit)
        cls._on_finish_methods.append(cls.update_ratelimit)

        for ratelimit_conf in ratelimit if isinstance(ratelimit, (list, tuple)) else [ratelimit]:
            cls._setup_ratelimit(ratelimit_conf, ratelimit_app_conf)

    @classmethod
    def _setup_ratelimit(cls, ratelimit, ratelimit_app_conf):
        for key in ('keys', 'limit'):
            if key not in ratelimit:
                raise ValueError(f'url:{cls.name}.ratelimit.{key}: missing')

        # All info for current ratelimit is in _ratelimit, which is added to cls._ratelimit
        _ratelimit = AttrDict(
            # Default the pool name to `pattern:`
            pool=ratelimit.get('pool', cls.conf.pattern),
            # Pick up the store location from the default app config
            store=cls._get_store(ratelimit_app_conf),
            key_fn=[],
        )
        cls._ratelimit.append(_ratelimit)

        # Convert keys: into list
        keys_spec = ratelimit['keys']
        # keys: daily, user => keys: [daily, user]
        if isinstance(keys_spec, str):
            keys_spec = cls.get_list(keys_spec, key=cls.name, eg='daily, user', caps=False)
        # keys: {function: ...} => keys: [{function: ...}]
        elif isinstance(keys_spec, dict):
            keys_spec = [keys_spec]
        # keys: must be a list
        elif not isinstance(keys_spec, (list, tuple)):
            raise ValueError(f'url:{cls.name}.ratelimit.keys: needs dict list, not {keys_spec}')

        # Pre-compile keys: into _ratelimit.key_fn = [key_fn, key_fn, ...]
        #   key_fn['function'](self) will return nth key
        #   key_fn['expiry'](self) will return nth expiry (in seconds)
        for index, key_spec in enumerate(keys_spec):
            if isinstance(key_spec, str):
                # Look up string keys like daily to predefined_keys.
                if key_spec in _PREDEFINED_KEYS:
                    key_spec = _PREDEFINED_KEYS[key_spec]
                # Or construct functions for `user.id`, etc
                else:
                    try:
                        key_spec = {'function': handler_expr(key_spec)}
                    except ValueError:
                        raise ValueError(f'url:{cls.name}.ratelimit.keys: {key_spec} is unknown')
            # {function: ...} or {key: ...} MUST be defined for a key. {expiry: ... } is optional
            if isinstance(key_spec, dict) and ('function' in key_spec or 'key' in key_spec):
                # Compile key/expiry functions into _ratelimit.key_fn[index]['function' / 'expiry']
                key_fn = {}
                for fn in ('function', 'expiry', 'key'):
                    if fn in key_spec:
                        if callable(key_spec[fn]):
                            key_fn[fn] = key_spec[fn]
                        else:
                            key_fn[fn] = build_transform(
                                {'function': key_spec[fn]},
                                vars={'handler': None},
                                filename=f'url:{cls.name}.ratelimit.keys[{index}].{fn}',
                                iter=False,
                            )
                _ratelimit.key_fn.append(key_fn)
            else:
                raise ValueError(f'url:{cls.name}.ratelimit.keys: {key_spec} has no function/key:')

        # Ensure limit: is a number or a {function: ...}
        limit_spec = ratelimit['limit']
        if isinstance(limit_spec, (int, float)):
            limit_spec = {'function': limit_spec}
        elif not isinstance(ratelimit['limit'], dict) or 'function' not in ratelimit['limit']:
            example = "{'function': number}"
            raise ValueError(f'url:{cls.name}.ratelimit.limit: needs {example}, not {limit_spec}')

        # Pre-compile limit: into _ratelimit.limit_fn
        _ratelimit.limit_fn = build_transform(
            limit_spec,
            vars={'handler': None},
            filename=f'url:{cls.name}.ratelimit.limit',
            iter=False,
        )

    @classmethod
    def reset_ratelimit(cls, pool: str, keys: List[Any], value: int = 0) -> bool:
        '''Reset the rate limit usage for a specific pool.

        Examples:
            >>> reset_ratelimit('/api', ['2022-01-01', 'x@example.org'])
            >>> reset_ratelimit('/api', ['2022-01-01', 'x@example.org'], 10)

        Parameters:
            pool: Rate limit pool to use. This is the url's `pattern:` unless you specified a
                `kwargs.ratelimit.pool:`
            keys: specific instance to reset. If your `ratelimit.keys` is `[daily, user.id]`,
                keys might look like `['2022-01-01', 'x@example.org']` to clear for that day/user
            value: sets the usage counter to this number (default: `0`)
        '''
        store = cls._get_store(conf.app.get('ratelimit'))
        key = json.dumps([pool] + keys)
        val = store.load(key, None)
        if val is not None and 'n' in val:
            val['n'] = value
            store.dump(key, val)
        else:
            return False

    @classmethod
    def setup_redirect(cls, redirect):
        '''
        Any handler can have a `redirect:` kwarg that looks like this:

        ```yaml
        redirect:
            query: next         # If the URL has a ?next=..., redirect to that page next
            header: X-Next      # Else if the header has an X-Next=... redirect to that
            url: ...            # Else redirect to this URL
        ```

        Only these 3 keys are allowed. All are optional, and checked in the
        order specified. So, for example:

        ```yaml
        redirect:
            header: X-Next      # Checks the X-Next header first
            query: next         # If it's missing, uses the ?next=
        ```

        You can also specify a string for redirect. `redirect: ...` is the same
        as `redirect: {url: ...}`.

        When any BaseHandler subclass calls `self.save_redirect_page()`, it
        stores the redirect URL in `session['_next_url']`. The URL is
        calculated relative to the handler's URL.

        After that, when the subclass calls `self.redirect_next()`, it
        redirects to `session['_next_url']` and clears the value. (If the
        `_next_url` was not stored, we redirect to the home page `/`.)

        Only some handlers implement redirection. But they all implement it in
        this same consistent way.
        '''
        # Ensure that redirect is a dictionary before proceeding.
        if isinstance(redirect, str):
            redirect = {'url': redirect}
        if not isinstance(redirect, dict):
            app_log.error(f'url:{cls.name}.redirect must be a URL or a dict, not {redirect!r}')
            return

        cls.redirects = []
        add = cls.redirects.append
        for key, value in redirect.items():
            if key == 'query':
                add(lambda h, v=value: h.get_argument(v, None))
            elif key == 'header':
                add(lambda h, v=value: h.request.headers.get(v))
            elif key == 'url':
                add(lambda h, v=value: v)

        # redirect.external=False disallows external URLs
        if not redirect.get('external', False):

            def no_external(method):
                def redirect_method(handler):
                    next_uri = method(handler)
                    if next_uri is not None:
                        target = urlsplit(next_uri)
                        if not target.scheme and not target.netloc:
                            return next_uri
                        req = handler.request
                        if req.protocol == target.scheme and req.host == target.netloc:
                            return next_uri
                        app_log.error(f'Not redirecting to external url: {next_uri}')

                return redirect_method

            cls.redirects = [no_external(method) for method in cls.redirects]

    @classmethod
    def setup_auth(cls, auth: Union[None, bool, dict]):
        # auth: if there's no auth: in handler, default to app.auth
        if auth is None:
            auth = conf.app.get('auth')
        # Treat True as an empty dict, i.e. auth: {}
        if auth is True:
            auth = AttrDict()
        # Set up the auth
        if isinstance(auth, dict):
            cls._auth = auth
            cls._auth_methods = cls.get_list(
                auth.get('methods', ''), 'auth.methods', '[GET, POST, OPTIONS]'
            )
            cls._on_init_methods.append(cls.authorize)
            cls.permissions = []
            # Add check for condition
            if auth.get('condition'):
                cls.permissions.append(
                    build_transform(
                        auth['condition'],
                        vars={'handler': None},
                        filename=f'url:{cls.name}.auth.permission',
                    )
                )
            # Add check for membership
            memberships = auth.get('membership', [])
            if not isinstance(memberships, list):
                memberships = [memberships]
            if len(memberships):
                cls.permissions.append(check_membership(memberships))
        elif auth:
            app_log.error(f'url:{cls.name}.auth is not a dict')

    def authorize(self):
        '''BaseMixin assumes every handler has an authorize() function'''
        pass

    @classmethod
    def setup_log(cls):
        '''
        Logs access requests to gramex.requests as a CSV file.
        '''
        logger = logging.getLogger('gramex.requests')
        keys = objectpath(conf, 'log.handlers.requests.keys', [])
        log_info = build_log_info(keys)
        cls.log_request = lambda handler: logger.info(log_info(handler))

    @classmethod
    def _error_fn(cls, error_code, error_config):
        template_kwargs = {}
        if 'autoescape' in error_config:
            if not error_config['autoescape']:
                template_kwargs['autoescape'] = None
            else:
                app_log.error(f'url:{cls.name}.error.{error_code}.autoescape can only be false')
        if 'whitespace' in error_config:
            template_kwargs['whitespace'] = error_config['whitespace']

        def error(*args, **kwargs):
            tmpl = gramex.cache.open(error_config['path'], 'template', **template_kwargs)
            return tmpl.generate(*args, **kwargs)

        return error

    @classmethod
    def setup_error(cls, error):
        '''
        Sample configuration:

        ```yaml
        error:
            404:
                path: template.json         # Use a template
                autoescape: false           # with no autoescape
                whitespace: single          # as a single line
                headers:
                    Content-Type: application/json
            500:
                function: module.fn
                args: [=status_code, =kwargs, =handler]
        ```
        '''
        if not error:
            return
        if not isinstance(error, dict):
            return app_log.error(f'url:{cls.name}.error is not a dict')
        # Compile all errors handlers
        cls.error = {}
        for error_code, error_config in error.items():
            try:
                error_code = int(error_code)
                if error_code < 100 or error_code > 1000:
                    raise ValueError()
            except ValueError:
                app_log.error(f'url.{cls.name}.error code {error_code} is not a number (100-1000)')
                continue
            if not isinstance(error_config, dict):
                return app_log.error(f'url:{cls.name}.error.{error_code} is not a dict')
            # Make a copy of the original. When we add headers, etc, it shouldn't affect original
            error_config = AttrDict(error_config)
            error_path, error_function = error_config.get('path'), error_config.get('function')
            if error_function:
                if error_path:
                    error_config.pop('path')
                    app_log.warning(
                        f'url.{cls.name}.error.{error_code} has function:. Ignoring path:'
                    )
                cls.error[error_code] = {
                    'function': build_transform(
                        error_config,
                        vars={'status_code': None, 'kwargs': None, 'handler': None},
                        filename=f'url:{cls.name}.error.{error_code}',
                    )
                }
            elif error_path:
                encoding = error_config.get('encoding', 'utf-8')
                cls.error[error_code] = {'function': cls._error_fn(error_code, error_config)}
                mime_type, encoding = mimetypes.guess_type(error_path, strict=False)
                if mime_type:
                    error_config.setdefault('headers', {}).setdefault('Content-Type', mime_type)
            else:
                app_log.error(f'url.{cls.name}.error.{error_code} must have path: or function:')
            # Add the error configuration for reference
            if error_code in cls.error:
                cls.error[error_code]['conf'] = error_config
        cls._write_error, cls.write_error = cls.write_error, cls._write_custom_error

    @classmethod
    def setup_xsrf(cls, xsrf_cookies):
        '''
        Sample configuration:

        ```yaml
        xsrf_cookies: false         # Disables xsrf_cookies
        xsrf_cookies: true          # or anything other than false keeps it enabled
        ```
        '''
        cls.check_xsrf_cookie = cls.noop if xsrf_cookies is False else cls.xsrf_ajax

    def xsrf_check_required(self):
        '''Returns True if the request is (likely) from a browser and needs XSRF check.

        This is used to handle XSRF. If the request is NOT from a browser (e.g. server, AJAX),
        no AJAX checks are required. If any of the following are true, it's not a browser request.

        1. `X-Requested-With: XMLHttpRequest`. XMLHttpRequest sends this
        2. `Sec-Fetch-Mode: cors`. [Fetch sends these][H-Fetch]

        [H-Fetch]: https://developer.mozilla.org/en-US/docs/Glossary/Fetch_metadata_request_header
        [H-Origin]: https://developer.mozilla.org/en-US/docs/Web/API/Request/mode
        '''
        return not (
            self.request.headers.get('X-Requested-With', '').lower() == 'xmlhttprequest'
            or self.request.headers.get('Sec-Fetch-Mode', '').lower() == 'cors'
        )

    def xsrf_ajax(self):
        '''Validates XSRF cookies if it's a browser-request (not AJAX)

        Internally, it uses Tornado's check_xsrf_cookie().
        '''
        if self.xsrf_check_required():
            return super(BaseHandler, self).check_xsrf_cookie()

    def noop(self):
        '''Does nothing. Used when overriding functions or providing a dummy operation'''
        pass

    def save_redirect_page(self):
        '''
        Loop through all redirect: methods and save the first available redirect
        page against the session. Defaults to previously set value, else `/`.

        See [setup_redirect][gramex.handlers.BaseMixin.setup_redirect].
        '''
        for method in self.redirects:
            next_url = method(self)
            if next_url:
                self.session['_next_url'] = urljoin(self.xrequest_uri, next_url)
                return
        self.session.setdefault('_next_url', '/')

    def redirect_next(self):
        '''
        Redirect the user `session['_next_url']`. If it does not exist,
        set it up first. Then redirect.

        See [setup_redirect][gramex.handlers.BaseMixin.setup_redirect].
        '''
        if '_next_url' not in self.session:
            self.save_redirect_page()
        self.redirect(self.session.pop('_next_url', '/'))

    @tornado.gen.coroutine
    def _cached_get(self, *args, **kwargs):
        cached = self.cachefile.get()
        if cached is not None:
            self.set_status(cached['status'])
            self._write_headers(cached['headers'])
            self.write(cached['body'])
        else:
            self.cachefile.wrap(self)
            yield self.original_get(*args, **kwargs)

    def _write_headers(self, headers):
        '''Write headers from a list of pairs that may be duplicated'''
        headers_written = set()
        for name, value in headers:
            # If value is explicitly False or None, clear header.
            # This gives a way to clear pre-set headers like the Server header
            if value is False or value is None:
                self.clear_header(name)
            elif name in headers_written:
                self.add_header(name, value)
            else:
                self.set_header(name, value)
                headers_written.add(name)

    def debug_exception(self, typ, value, tb):
        super(BaseHandler, self).log_exception(typ, value, tb)
        import pdb  # noqa T100

        pdb.post_mortem(tb)

    def _write_custom_error(self, status_code, **kwargs):
        if status_code in self.error:
            try:
                result = self.error[status_code]['function'](
                    status_code=status_code, kwargs=kwargs, handler=self
                )
                headers = self.error[status_code].get('conf', {}).get('headers', {})
                self._write_headers(headers.items())
                # result may be a generator / list from build_transform,
                # or a str/bytes/unicode from Template.generate. Handle both
                if isinstance(result, (str, bytes)):
                    self.write(result)
                else:
                    for item in result:
                        self.write(item)
                return
            except Exception:
                app_log.exception(f'url:{self.name}.error.{status_code} raised an exception')
        # HTTP 429 error code reports ratelimits. Set the headers if ratelimit is set.
        # Note: Previous headers are cleared, so we need to explicitly write them here.
        if status_code == TOO_MANY_REQUESTS and hasattr(self, '_ratelimit'):
            self.set_ratelimit_headers()
        # If error was not written, use the default error
        self._write_error(status_code, **kwargs)

    @property
    def session(self):
        '''
        By default, session is not implemented. You need to specify a
        `session:` section in `gramex.yaml` to activate it. It is replaced by
        the `get_session` method as a property.
        '''
        raise NotImplementedError('Specify a session: section in gramex.yaml')

    def _set_new_session_id(self, expires_days):
        '''Sets a new random session ID as the sid: cookie. Returns a bytes object'''
        session_id = b2a_base64(os.urandom(24))[:-1]
        kwargs = dict(self._session_cookie)
        kwargs['expires_days'] = expires_days
        # Treat expiry: false as None, since expiry: null won't set null, it clears YAML
        if kwargs['expires_days'] is False:
            kwargs['expires_days'] = None
        # Use Secure cookies on HTTPS to prevent leakage into HTTP
        if self.request.protocol == 'https':
            kwargs['secure'] = True
        # Websockets cannot set cookies. They raise a RuntimeError. Ignore those.
        with contextlib.suppress(RuntimeError):
            self.set_secure_cookie(self._session_cookie_id, session_id, **kwargs)
        # Warn if app.session.domain is x.com but request comes from y.com.
        host = self.request.host_name
        if (
            'domain' in kwargs
            and not host.endswith(kwargs['domain'])
            and not kwargs['domain'].endswith('.local')
        ):
            app_log.warning(
                f'{self.name}: session.domain={kwargs["domain"]} but cookie sent to {host}'
            )
        return session_id

    def get_session(self, expires_days=None, new=False):
        '''
        Return the session object for the cookie "sid" value.
        If no "sid" cookie exists, set up a new one.
        If no session object exists for the sid, create it.
        By default, the session object contains a "id" holding the "sid" value.

        The session is a dict. You must ensure that it is JSON serializable.

        Sessions use these pre-defined timing keys (values are timestamps):

        - `_t` is the expiry time of the session
        - `_l` is the last time the user accessed a page. Updated by
          [setup_redirect][gramex.handlers.BaseMixin.set_last_visited]
        - `_i` is the inactive expiry duration in seconds, i.e. if `now > _l +
          _i`, the session has expired.

        `new=` creates a new session to avoid session fixation.
        https://www.owasp.org/index.php/Session_fixation.
        [`set_user()`][gramex.handlers.AuthHandler.set_user] uses it.
        When the user logs in:

        - If no old session exists, it returns a new session object.
        - If an old session exists, it creates a new "sid" and new session
          object, copying all old contents, but updates the "id" and expiry (_t).
        '''
        if expires_days is None:
            expires_days = self._session_expiry
        # If the expiry time is None, keep in the session store for 1 day
        store_expires = time.time() + (1 if expires_days is None else expires_days) * 24 * 60 * 60
        created_new_sid = False
        if getattr(self, '_session', None) is None:
            # Populate self._session based on the sid. If there's no sid cookie,
            # generate one and create an associated session object
            session_id = self.get_secure_cookie(self._session_cookie_id, max_age_days=9999999)
            # If there's no session id cookie "sid", create a random 32-char cookie
            if session_id is None:
                session_id = self._set_new_session_id(expires_days)
                created_new_sid = True
            # Convert bytes session to unicode before using
            session_id = session_id.decode('ascii')
            # If there's no stored session associated with it, create it
            self._session = self._session_store.load(session_id, {'_t': store_expires})
            # Overwrite id to the session ID even if a handler has changed it
            self._session['id'] = session_id
        # At this point, the "sid" cookie and self._session exist and are synced
        s = self._session
        old_sid = s['id']
        # If session has expiry keys _i and _l defined, check for expiry. Not otherwise
        if '_i' in s and '_l' in s and time.time() > s['_l'] + s['_i']:
            new = True
            s.clear()
        if new and not created_new_sid:
            new_sid = self._set_new_session_id(expires_days).decode('ascii')
            # Update expiry and new SID on session
            s.update(id=new_sid, _t=store_expires)
            # Delete old contents. No _t also means expired
            self._session_store.dump(old_sid, {})

        return s

    def save_session(self):
        '''Persist the session object as a JSON'''
        if getattr(self, '_session', None) is not None:
            self._session_store.dump(self._session['id'], self._session)

    def otp(
        self,
        expire: float = 60,
        user: Union[str, dict] = None,
        size: int = None,
        type: str = 'OTP',
        reset: bool = False,
        **kwargs: Dict[str, Any],
    ) -> str:
        '''Return one-time password valid for `expire` seconds.

        The OTP is used as the X-Gramex-OTP header or in `?gramex-otp=` on any request.
        This overrides the user with the passed `user` object for that session.

        Parameters:
            expire: Time when this token expires, in seconds (e.g. `60` means 1 minute from now)
            user: User object to store against token. Defaults to current user. Raises HTTP 403
                Unauthorized if there's no user
            size: Length of the OTP in characters. `None` means a full hash string
            type: Identifier for type of OTP. `OTP` for OTPs. Use `Key` for API keys. Auth handlers
                use their class names, e.g. `DBAuth`, `SMSAuth`, `EMailAuth`.
            reset: True to delete all previous OTPs for the user
            kwargs: Additional columns to add (if they exist)

        Returns:
            Generated OTP

        Internally, this stores it in `storelocations.otp` database in a table with 4+ keys:

        1. `token`: Generated OTP with `size` characters
        2. `user`: The passed `user` string or dict, JSON-encoded
        3. `type`: The passed `type` string, stored as is
        4. `expire`: The expiry time in seconds since epoch
        5. Any additional columns passed in `kwargs`, if they exist
        '''
        user = self.current_user if user is None else user
        if not user:
            raise HTTPError(UNAUTHORIZED)
        from uuid import uuid4

        user_obj = json.dumps(user)
        if reset:
            gramex.data.delete(**gramex.service.storelocations.otp, args={'user': [user_obj]})

        otp = uuid4().hex[:size]
        gramex.data.insert(
            **gramex.service.storelocations.otp,
            args={
                'token': [otp],
                'user': [user_obj],
                'type': [type],
                'expire': [time.time() + expire],
                **{key: [val] for key, val in kwargs.items()},
            },
        )
        return otp

    def get_otp(self, key: str, revoke: bool = False) -> Union[str, dict, None]:
        '''Return the user object given the OTP key. Revoke the OTP if requested.

        Parameters:
            key: OTP to return
            revoke: True to revoke the OTP. False to retain it

        Returns:
            `None` if the OTP `key` doesn't exist or has expired.
                Else a dict with keys `user`, `expire`, `type` and `token`.
        '''
        rows = gramex.data.filter(**gramex.service.storelocations.otp, args={'token': [key]})
        if len(rows) == 0:
            return None
        row = rows.iloc[0].to_dict()
        if revoke:
            gramex.data.delete(
                **gramex.service.storelocations.otp, id=['token'], args={'token': [key]}
            )
        # Skip expired tokens
        if row['expire'] <= time.time():
            return None
        # Return the user column parsed as JSON.
        # Add custom keys from the table to the user object. Don't overwrite NULL values.
        row['user'] = json.loads(row['user'])
        custom_keys = [key for key in row if key not in {'user', 'token', 'expire'}]
        if isinstance(row['user'], dict):
            for key in custom_keys:
                row['user'][key] = row['user'].get(key, None) if row[key] is None else row[key]
        else:
            app_log.warning('Cannot add custom keys to non-dict "user" in: %r', row)
        return row

    def revoke_otp(self, key: str) -> Union[str, dict, None]:
        '''Revoke an OTP. Returns the user object from [gramex.handlers.BaseMixin.get_otp][].'''
        return self.get_otp(key, revoke=True)

    def apikey(
        self,
        expire: float = 1e9,
        user: Union[str, dict] = None,
        size: int = None,
        **kwargs: Dict[str, Any],
    ) -> str:
        '''Return API Key. Usage is same as [gramex.handlers.BaseMixin.otp][]

        The API key is used as the X-Gramex-Key header or in `?gramex-key=` on any request.
        This overrides the user with the passed `user` object for that session.
        '''
        return self.otp(expire=expire, user=user, size=size, type='Key', **kwargs)

    def revoke_apikey(self, key: str) -> Union[str, dict, None]:
        '''Revoke API Key. Returns the user object from [gramex.handlers.BaseMixin.get_otp][].'''
        return self.revoke_otp(key)

    def override_user(self):
        '''Internal method to override the user.

        Use `X-Gramex-User` HTTP header to override current user for the session.
        Use `X-Gramex-OTP` HTTP header to set user based on OTP, or `?gramex-otp=`.
        Use `X-Gramex-Key` HTTP header to set user based on API key, or `?gramex-key=`.
        '''
        headers = self.request.headers
        cipher = headers.get('X-Gramex-User')
        if cipher:
            try:
                user = json.loads(
                    decode_signed_value(
                        conf.app.settings['cookie_secret'],
                        'user',
                        cipher,
                        max_age_days=self._session_expiry,
                    )
                )
            except Exception:
                raise HTTPError(BAD_REQUEST, f'{self.name}: invalid X-Gramex-User: {cipher}')
            else:
                app_log.debug(f'{self.name}: Overriding user to {user!r}')
                self.session['user'] = user
                return
        # OTP is specified as an X-Gramex-OTP header or ?gramex-otp argument.
        # API Key is specified as an X-Gramex-Key header or ?gramex-key argument.
        # Override the user if either is specified.
        for key in ('OTP', 'Key'):
            token = headers.get(f'X-Gramex-{key}') or self.get_argument(
                f'gramex-{key.lower()}', None
            )
            if token:
                # Revoke OTP keys. Don't revoke API keys
                row = self.get_otp(token, revoke=key == 'OTP')
                if not row:
                    raise HTTPError(
                        BAD_REQUEST, f'{self.name}: invalid/expired Gramex {key}: {token}'
                    )
                self.session['user'] = row['user']

    def set_last_visited(self):
        '''Update session last visited time if we track inactive expiry.

        - `session._l` is the last time the user accessed a page.
        - `session._i` is the seconds of inactivity after which the session expires.
        - If `session._i` is set (we track inactive expiry), we set `session._l` to now.
        '''
        # Called by BaseHandler.prepare() when any user accesses a page.
        # For efficiency reasons, don't call get_session every time. Check
        # session only if there's a valid sid cookie (with possibly long expiry)
        if self.get_secure_cookie(self._session_cookie_id, max_age_days=9999999):
            session = self.get_session()
            if '_i' in session:
                session['_l'] = time.time()

    def check_ratelimit(self):
        '''Raise HTTP 429 if usage exceeds rate limit. Set X-Ratelimit-* HTTP headers'''
        for ratelimit in self._ratelimit:
            # If no expiry is specified, store for 100 years
            expiries = [3155760000]
            # Get the rate limit key, limit and expiry
            keys = [ratelimit.pool]
            for key_fn in ratelimit.key_fn:
                if 'key' in key_fn:
                    predefined_key = key_fn['key'](self)
                    if predefined_key in _PREDEFINED_KEYS:
                        keys.append(_PREDEFINED_KEYS[predefined_key]['function'](self))
                        expiries.append(_PREDEFINED_KEYS[predefined_key]['expiry'](self))
                if 'function' in key_fn:
                    keys.append(key_fn['function'](self))
                if 'expiry' in key_fn:
                    expiries.append(key_fn['expiry'](self))

            ratelimit.key = json.dumps(keys)
            # Note: if ratelimit_fn() returns a non-int, check_ratelimit will fail. Let it fail.
            ratelimit.limit = ratelimit.limit_fn(self)
            ratelimit.expiry = min(expiries)

            # Ensure usage does not hit limit
            ratelimit.usage = ratelimit.store.load(ratelimit.key, {'n': 0}).get('n', 0)
            if ratelimit.usage >= ratelimit.limit:
                raise HTTPError(
                    TOO_MANY_REQUESTS,
                    f'{ratelimit.pool}: {ratelimit.key} hit rate limit {ratelimit.limit}',
                )
        self.set_ratelimit_headers()

    def update_ratelimit(self):
        '''If request succeeds, increase rate limit usage count by 1'''
        # If response is a HTTP error, don't count towards rate limit
        if self.get_status() >= 400:
            return
        for ratelimit in self._ratelimit:
            # If check_ratelimit failed (e.g. invalid function) and didn't set a key, skip update
            if 'key' not in ratelimit:
                return
            # Increment the rate limit by 1
            usage_obj = ratelimit.store.load(ratelimit.key, {'n': 0})
            usage_obj['n'] += 1
            usage_obj['_t'] = time.time() + ratelimit.expiry
            ratelimit.store.dump(ratelimit.key, usage_obj)

    def get_ratelimit(self):
        '''Get the rate limit with the least remaining usage for the current request.

        If there are multiple rate limits, it picks the one with least remaining usage and
        returns an AttrDict with these keys:

        - `limit`: the limit on the rate limit (e.g. 3)
        - `usage`: the usage so far (BEFORE current request, e.g. 0, 1, 2, 3, ...)
        - `remaining`: the remaining requests (AFTER current request, e.g. 2, 1, 0, 0, ...)
        - `expiry`: seconds to expiry for this rate limit
        '''
        ratelimit = min(self._ratelimit, key=lambda r: r.limit - r.usage)
        return AttrDict(
            limit=ratelimit.limit,
            usage=ratelimit.usage,
            # ratelimit.usage goes 0, 1, 2, ...
            # If limit is 3, remaining goes 2, 1, 0, ... -- use (limit - usage - 1)
            # But when usage hits 3, don't show remaining = -1. Show remaining = 0 using max()
            remaining=max(ratelimit.limit - ratelimit.usage - 1, 0),
            expiry=ratelimit.expiry,
        )

    def set_ratelimit_headers(self):
        '''Sets the headers from the [Ratelimit HTTP headers draft][1].

        - `X-Ratelimit-Limit` is the rate limit
        - `X-Ratelimit-Remaining` has the remaining requests
        - `X-Ratelimit-Reset` has seconds after which the rate limit resets
        - `Retry-After` is same as X-Ratelimit-Reset, but is set only if the limit is exceeded

        [1]: https://www.ietf.org/archive/id/draft-polli-ratelimit-headers-02.html
        '''
        # Pick the rate limit with the lowest remaining count
        ratelimit = self.get_ratelimit()
        self.set_header('X-Ratelimit-Limit', str(ratelimit.limit))
        self.set_header('X-Ratelimit-Remaining', str(ratelimit.remaining))
        self.set_header('X-RateLimit-Reset', str(ratelimit.expiry))
        if ratelimit.usage >= ratelimit.limit:
            self.set_header('Retry-After', str(ratelimit.expiry))

    def initialize_handler(self):
        '''Initialize self.args and other handler attributes'''
        # self.request.arguments does not handle unicode keys well. It returns latin-1 unicode.
        # https://github.com/tornadoweb/tornado/issues/2733
        # Convert this to proper unicode using UTF-8 and store in self.args
        self.args = {}
        for k in self.request.arguments:
            key = k.encode('latin-1').decode('utf-8')
            # Invalid unicode (e.g. ?x=%f4) throws HTTPError. This disrupts even
            # error handlers. So if there's invalid unicode, log & continue.
            try:
                self.args[key] = self.get_arguments(k)
            except HTTPError:
                app_log.exception(f'{self.name}: Invalid unicode ?{k}')
        self._session, self._session_json = None, 'null'
        if self.cache:
            self.cachefile = self.cache()
            self.original_get = self.get
            self.get = self._cached_get
        if self._set_xsrf:
            self.xsrf_token

    def update_body_args(self):
        '''Update self.args with JSON body if Content-Type is application/json'''
        if self.request.body:
            content_type = self.request.headers.get('Content-Type', '')
            if content_type == 'application/json':
                self.args.update(json.loads(self.request.body))

    def get_arg(self, name, default=..., first=False):
        '''
        Returns the value of the argument with the given name. Similar to
        `.get_argument` but uses `self.args` instead.

        If default is not provided, the argument is considered to be
        required, and we raise a `MissingArgumentError` if it is missing.

        If the argument is repeated, we return the last value. If `first=True`
        is passed, we return the first value.

        `self.args` is always UTF-8 decoded unicode. Whitespaces are stripped.
        '''
        if name not in self.args:
            if default is ...:
                raise MissingArgumentError(name)
            return default
        return self.args[name][0 if first else -1]

    @classmethod
    def setup_validate(cls, validate: Union[list, dict, str, None]):
        if validate is None:
            return
        cls._validate = []
        validate = validate if isinstance(validate, (list, tuple)) else [validate]
        for index, rule in enumerate(validate):
            rule = dict(rule) if isinstance(rule, dict) else {'function': rule}
            code = rule.pop('code', BAD_REQUEST)
            reason = rule.pop('reason', f'{cls.name}: validate: {repr(rule)} failed')
            fn = build_transform(
                rule, {'handler': None}, filename=f'{cls.name}.validate[{index}]', iter=False
            )
            cls._validate.append({'function': fn, 'code': code, 'reason': reason, 'rule': rule})
        if len(validate):
            cls._on_init_methods.append(cls.check_validate)

    def check_validate(self):
        for rule in self._validate:
            try:
                result = rule['function'](self)
            except Exception:
                app_log.exception(rule['reason'])
                raise HTTPError(rule['code'], reason=rule['reason'])
            if not result:
                raise HTTPError(rule['code'], reason=rule['reason'])


class BaseHandler(RequestHandler, BaseMixin):
    '''
    BaseHandler provides auth, caching and other services common to all request
    handlers. All RequestHandlers must inherit from BaseHandler.
    '''

    def initialize(self, **kwargs):
        self.initialize_handler()

        # Set the method to the ?x-http-method-overrride argument or the
        # X-HTTP-Method-Override header if they exist
        if 'x-http-method-override' in self.args:
            self.request.method = self.args.pop('x-http-method-override')[0].upper()
        elif 'X-HTTP-Method-Override' in self.request.headers:
            self.request.method = self.request.headers['X-HTTP-Method-Override'].upper()

    def prepare(self):
        # If X-Request-URI is specified, use it. Else, when redirecting use RELATIVE URL. That
        # allows nginx to proxy_redirect Location headers
        self.xrequest_uri = self.request.headers.get('X-Request-URI', self.request.uri)
        # When passing to URL query parameters (e.g. /login/?next=), use the full URL
        self.xrequest_full_url = urljoin(self.request.full_url(), self.xrequest_uri)
        # For third-party redirection (e.g. Google Auth / Twitter needs a callback URI), then use
        # the full URL WITHOUT query parameters
        self.xredirect_uri = '{0.scheme:s}://{0.netloc:s}{0.path:s}'.format(
            urlsplit(self.xrequest_full_url)
        )
        # If X-Gramex-Root is specified, treat that as the application's root URL
        self.gramex_root = self.request.headers.get('X-Gramex-Root', '').rstrip('/')
        for method in self._on_init_methods:
            method(self)

    def set_default_headers(self):
        # Only set BaseHandler headers.
        # Don't set headers for the specific class. Those are overrides handled
        # by the respective classes, not the default headers.
        headers = [('Server', server_header)]
        headers += list(objectpath(conf, 'handlers.BaseHandler.headers', {}).items())
        self._write_headers(headers)

    def on_finish(self):
        # Loop through class-level callbacks
        for callback in self._on_finish_methods:
            callback(self)

    def get_current_user(self):
        '''Return the `user` key from the session as an AttrDict if it exists.'''
        result = self.session.get('user')
        return AttrDict(result) if isinstance(result, dict) else result

    def log_exception(self, typ, value, tb):
        '''Store the exception value for logging'''
        super(BaseHandler, self).log_exception(typ, value, tb)
        # _exception is stored for use by log_request. Sample error string:
        # ZeroDivisionError: integer division or modulo by zero
        self._exception = traceback.format_exception_only(typ, value)[0].strip()

    def authorize(self):
        # If specific methods are mentioned, authorize only if a mentioned method is used
        auth_methods = getattr(self, '_auth_methods', None)
        if auth_methods and self.request.method not in auth_methods:
            return
        # If CORS auth is specified, don't authorize for OPTIONS (pre-flight request)
        if self.request.method == 'OPTIONS' and getattr(self, '_cors', {}).get('auth'):
            return
        if not self.current_user:
            # Redirect browser GET/HEAD requests to login URL (if it's a string)
            if self.xsrf_check_required() and self.request.method in ('GET', 'HEAD'):
                auth = getattr(self, '_auth', {})
                url = auth.get('login_url', self.gramex_root + self.get_login_url())
                # If login_url: false, don't redirect to a login URL. Only redirect if it's a URL
                if isinstance(url, str):
                    # Redirect to the login_url adding ?next=<X-Request-URI>
                    p = urlsplit(url)
                    q = parse_qsl(p.query)
                    q.append((auth.get('query', 'next'), self.xrequest_full_url))
                    target = urlunsplit((p.scheme, p.netloc, p.path, urlencode(q), p.fragment))
                    self.redirect(target)
                    return
            # Else, send a 401 header
            raise HTTPError(UNAUTHORIZED)

        # If the user doesn't have permissions, show 403 (with template)
        for permit_generator in self.permissions:
            # If any permit_generator raises an Exception, let it bubble up to a HTTP 500
            for result in permit_generator(self):
                if not result:
                    template = self.conf.kwargs.auth.get('template')
                    if template:
                        self.set_status(FORBIDDEN)
                        self.render(template)
                    raise HTTPError(FORBIDDEN)

    def argparse(self, *args, **kwargs):
        '''
        Parse URL query parameters and return an AttrDict. For example:

        ```python
        args = handler.argparse('x', 'y')
        args.x      # is the last value of ?x=value
        args.y      # is the last value of ?y=value
        ```

        A missing `?x=` or `?y=` raises a HTTP 400 error mentioning the
        missing key.

        For optional arguments, use:

        ```python
        args = handler.argparse(z={'default': ''})
        args.z      # returns '' if ?z= is missing
        ```

        You can convert the value to a type:

        ```python
        args = handler.argparse(limit={'type': int, 'default': 100})
        args.limit      # returns ?limit= as an integer
        ```

        You can restrict the choice of values. If the query parameter is not in
        choices, we raise a HTTP 400 error mentioning the invalid key & value:

        ```python
        args = handler.argparse(gender={'choices': ['M', 'F']})
        args.gender      # returns ?gender= which will be 'M' or 'F'
        ```

        You can retrieve multiple values as a list:

        ```python
        args = handler.argparse(cols={'nargs': '*', 'default': []})
        args.cols       # returns an array with all ?col= values
        ```

        `type:` conversion and `choices:` apply to each value in the list.

        To return all arguments as a list, pass `list` as the first parameter:

        ```python
        args = handler.argparse(list, 'x', 'y')
        args.x      # ?x=1 sets args.x to ['1'], not '1'
        args.y      # Similarly for ?y=1
        ```

        To handle unicode arguments and return all arguments as `str` or
        `unicode` or `bytes`, pass the type as the first parameter:

        ```python
        args = handler.argparse(str, 'x', 'y')
        args = handler.argparse(bytes, 'x', 'y')
        args = handler.argparse(unicode, 'x', 'y')
        ```

        By default, all arguments are added as str in PY3 and unicode in PY2.

        There are the full list of parameters you can pass to each keyword
        argument:

        - name: Name of the URL query parameter to read. Defaults to the key
        - required: Whether or not the query parameter may be omitted
        - default: The value produced if the argument is missing. Implies required=False
        - nargs: The number of parameters that should be returned. '*' or '+'
          return all values as a list.
        - type: Python type to which the parameter should be converted (e.g. `int`)
        - choices: A container of the allowable values for the argument (after type conversion)

        You can combine all these options. For example:

        ```python
        args = handler.argparse(
            'name',                         # Raise error if ?name= is missing
            department={'name': 'dept'},    # ?dept= is mapped to args.department
            org={'default': 'Gramener'},    # If ?org= is missing, defaults to Gramener
            age={'type': int},              # Convert ?age= to an integer
            married={'type': bool},         # Convert ?married to a boolean
            alias={'nargs': '*'},           # Convert all ?alias= to a list
            gender={'choices': ['M', 'F']}, # Raise error if gender is not M or F
        )
        ```
        '''
        result = AttrDict()

        args_type = str
        if len(args) > 0 and args[0] in (str, bytes, list, None):
            args_type, args = args[0], args[1:]

        for key in args:
            result[key] = self.get_argument(key, None)
            if result[key] is None:
                raise HTTPError(BAD_REQUEST, f'{key}: missing ?{key}=')
        for key, config in kwargs.items():
            name = config.get('name', key)
            val = self.args.get(name, [])

            # default: set if query is missing
            # required: check if query is defined at all
            if len(val) == 0:
                if 'default' in config:
                    result[key] = config['default']
                    continue
                if config.get('required', False):
                    raise HTTPError(BAD_REQUEST, f'{key}: missing ?{name}=')

            # nargs: select the subset of items
            nargs = config.get('nargs', None)
            if isinstance(nargs, int):
                val = val[:nargs]
                if len(val) < nargs:
                    val += [''] * (nargs - len(val))
            elif nargs not in ('*', '+', None):
                raise ValueError(f'{key}: invalid nargs {nargs}')

            # convert to specified type
            newtype = config.get('type', None)
            if newtype is not None:
                newval = []
                for v in val:
                    try:
                        newval.append(newtype(v))
                    except ValueError:
                        raise HTTPError(
                            BAD_REQUEST, f'{key}: type error ?{name}={v} to {newtype!r}'
                        )
                val = newval

            # choices: check valid items
            choices = config.get('choices', None)
            if isinstance(choices, (list, dict, set)):
                choices = set(choices)
                for v in val:
                    if v not in choices:
                        raise HTTPError(BAD_REQUEST, f'{key}: invalid choice ?{name}={v}')

            # Set the final value
            if nargs is None:
                if len(val) > 0:
                    result[key] = val[-1]
            else:
                result[key] = val

        # Parse remaining keys
        if args_type is list:
            for key, val in self.args.items():
                if key not in args and key not in kwargs:
                    result[key] = val
        elif args_type in (str, bytes):
            for key, val in self.args.items():
                if key not in args and key not in kwargs:
                    result[key] = args_type(val[0])

        return result

    def create_template_loader(self, template_path):
        settings = self.application.settings
        return CacheLoader(
            template_path,
            autoescape=settings['autoescape'],
            whitespace=settings.get('template_whitespace', None),
        )


class BaseWebSocketHandler(WebSocketHandler, BaseMixin):
    def initialize(self, **kwargs):
        self.initialize_handler()

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        for method in self._on_init_methods:
            method(self)
        yield super(BaseWebSocketHandler, self).get(*args, **kwargs)

    def on_close(self):
        # Loop through class-level callbacks
        for callback in self._on_finish_methods:
            callback(self)

    def get_current_user(self):
        '''Return the `user` key from the session as an AttrDict if it exists.'''
        result = self.session.get('user')
        return AttrDict(result) if isinstance(result, dict) else result

    def authorize(self):
        '''If a valid user isn't logged in, send a message and close connection'''
        if not self.current_user:
            raise HTTPError(UNAUTHORIZED)
        for permit_generator in self.permissions:
            for result in permit_generator(self):
                if not result:
                    raise HTTPError(FORBIDDEN)


class SetupFailedHandler(RequestHandler, BaseMixin):
    '''
    Reports that the setup() operation has failed.

    Used by gramex.services.init() when setting up URLs. If it's not able to set
    up a handler, it replaces it with this handler.
    '''

    def get(self):
        six.reraise(*self.exc_info)


def check_membership(memberships):
    '''
    Return a generator that checks all memberships for a user, and yields True if
    any membership is allowed, else False
    '''
    # Pre-process memberships into an array of {objectpath: set(values)}
    conds = [
        {
            keypath: set(values) if isinstance(values, list) else {values}
            for keypath, values in cond.items()
        }
        for cond in memberships
    ]

    def allowed(self):
        user = self.current_user
        for cond in conds:
            if _check_condition(cond, user):
                yield True
                break
        else:
            yield False

    return allowed


def _check_condition(condition, user):
    '''
    A condition is a dictionary of {keypath: values}. Extract the keypath from
    the user. Check if the value is in the values list. If not, this condition
    fails.
    '''
    for keypath, values in condition.items():
        node = objectpath(user, keypath)
        # If nothing exists at keypath, the check fails
        if node is None:
            return False
        # If the value is a list, it must overlap with values
        elif isinstance(node, list):
            if not set(node) & values:
                return False
        # If the value is not a list, it must be present in values
        elif node not in values:
            return False
    return True


_PREDEFINED_KEYS = {
    "hourly": time_key('%Y-%m-%d %H', freq='H'),
    "daily": time_key('%Y-%m-%d', offset='DateOffset'),
    "weekly": time_key('%Y %U', offset='Week'),
    "monthly": time_key('%Y-%m', offset='MonthBegin'),
    "yearly": time_key('%Y', offset='YearBegin'),
}
